/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"

// global defines :  RTAVR_INCLUDE_IOR, RTAVR_INCLDUE_RAM, RTAVR_INCLUDE_ROM

//
// Module Name: rtavr_alu
//   ver 001
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                                        all-in-one  
// Number of Slice Flip Flops:              xxx
// Number of 4 input LUTs:                  xxx
// Number of occupied Slices:               xxx
// Total Number of 4 input LUTs:            xxx
// Number of bonded IOBs :                  xxx
//   IOB Flip Flops :                       xxx
// Number of BUFGMUXs :                     xxx
// Number used for Dual Port RAMs           xxx
// Number of RAMB16BWEs :                   xxx
// Clock to Setup 
//on destination clock CLK2X              

module rtavr_alu (
		, input RR_IMM
		, input ALU_WC

		, input CMD_ALU
		, input CMD_ALU_WR, CMD_ALU_WRF
		, input CMD_ALU_SUB_ALL
//		, input CMD_ALU_MOV
		, input CMD_ALU_ADD
		, input CMD_ALU_OR
		, input CMD_ALU_EOR

		, input CMD_ALU_SUB
		, input CMD_ALU_SBC
		, input CMD_ALU_ADD
		, input CMD_ALU_Z
		, input CMD_ALU_V0
		, input CMD_ALU_NEG
		, input CMD_ALU_OP1

		, input [7:0] Rd_in
		, input [7:0] Rr_in
		, input [7:0] IMM_DATA
		, output [7:0] ALU_OUT
		);

  wire [7:0] ALU_RR = RR_IMM ? IMM_DATA : Rr_in;
  
  wire [8:0] ALU_OUT = CMD_ALU_ADD ? ({1'b0, Rd_in}  + {1'b0, ALU_RR }
				   + {8'b0, (ALU_WC)? C_in_l : 1'b0 })
		     : CMD_ALU_SUB_ALL ? ({1'b0,Rd_in} - {1'b0, ALU_RR }
				   - {8'b0, (ALU_WC) ? 1'b0: C_in_l })
		     : CMD_ALU_OR ?  {C_in_l, (Rd_in | ALU_RR) }
		     : CMD_ALU_EOR ? {C_in_l, (Rd_in ^ ALU_RR) }
		     : CMD_ALU_AND ? {C_in_l, (Rd_in & ALU_RR) }
		     : {C_in_l, ALU_RR } ;

  wire Rr_in_3  = ALU_RR[3];
  wire Rr_in_7  = ALU_RR[7];

//  wire [7:0] Rd_out = GPR_DI[7:0];
  wire [7:0] Rd_out = ALU_OUT[7:0];  // +12 slices

      assign C_out = CMD_ALU_WRF & ALU_OUT[8] | ~(CMD_ALU_WRF) & C_in ; 

      assign Z_out = ( CMD_ALU_SBC & Z_in | CMD_ALU_Z ) & (ALU_OUT[7:0] == 0) 
  		 | ~(CMD_ALU_SBC | CMD_ALU_Z ) & Z_in ;

      assign V_out = CMD_ALU_ADD & (Rd_out[7] & ~Rd_in[7] & ~Rr_in_7
                		| ~Rd_out[7] &  Rd_in[7] &  Rr_in_7 )
	      | CMD_ALU_SUB  & (   Rd_out[7] & ~Rd_in[7] &  Rr_in_7
               			| ~Rd_out[7] &  Rd_in[7] & ~Rr_in_7 )
	      | ~(CMD_ALU_ADD | CMD_ALU_SUB | CMD_ALU_V0) & V_in;

      assign H_out = CMD_ALU_ADD & ( 
			  ~Rd_out[3] & Rd_in[3] 
			| ~Rd_out[3] & Rr_in_3
			|  Rd_in[3]  & Rr_in_3 )
	       | CMD_ALU_SUB & (
			   Rd_out[3] & ~Rd_in[3] 
			|  Rd_out[3] &  Rr_in_3
			| ~Rd_in[3]  &  Rr_in_3 )
	       | ~(CMD_ALU_ADD | CMD_ALU_SUB) & H_in;

endmodule
