/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_TRI

//`define OUTPUT_MUX_TRI

`define TIM16_FUNC_MINIMUM

// Module Name: rtavr_ior_timer16
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
// BASE_ADDR+0: OCR_AL
// BASE_ADDR+1: OCR_AH
// BASE_ADDR+2: OCR_BL
// BASE_ADDR+3: OCR_BH
// BASE_ADDR+4: TCNT_L
// BASE_ADDR+5: TCNT_H
// BASE_ADDR+6: TCCR_B
// BASE_ADDR+7: TCCR_A
//
//
//
//                          DO:  NORMAL    TRISTATE(FAKE)   
// Number of Slice Flip Flops:     49         49   
// Number of 4 input LUTs:         69         73
// Number of occupied Slices:      56         57
// Total Number of 4 input LUTs:   78         82 
// Number of bonded IOBs :         28
//   IOB Flip Flops :              -
// Number of BUFGMUXs :             1
// Number of RAMB16BWEs :          -

module rtavr_ior_timer16(
		input CLK,
		input RESET,

		input [7:0] DI,
		output [7:0] DO

		, input WR,RD
		, input [5:0] ADDR

		// external I/O
		, output TIM16_OVF
		, output TIM16_COMPA	
		, output TIM16_COMPB

		, input [9:0] PS

		);
  parameter BASE_ADDR = 6'h15; // OCR0B 0x15
  wire [9:0] r_prescaler;
  reg  [7:0] r_tmp;
  reg  [15:0] r_tcnt;
  reg  [15:0] r_ocr_b;
  reg  [15:0] r_ocr_a;
  reg  r_tim_compa;
  reg  r_tim_compb;
  reg  r_tim_ovf;
 
  assign TIM16_OVF = r_tim_ovf;
  assign TIM16_COMPA = r_tim_compa;
  assign TIM16_COMPB = r_tim_compb;

  assign r_prescaler = PS;

  wire OE_ocr_bl = RD & (ADDR == BASE_ADDR);
  wire WE_ocr_bl = WR & (ADDR == BASE_ADDR);
  wire OE_ocr_bh = RD & (ADDR == BASE_ADDR + 1);
  wire WE_ocr_bh = WR & (ADDR == BASE_ADDR + 1);
  wire OE_ocr_al = RD & (ADDR == (BASE_ADDR + 2));
  wire WE_ocr_al = WR & (ADDR == (BASE_ADDR + 2));
  wire OE_ocr_ah = RD & (ADDR == (BASE_ADDR + 3));
  wire WE_ocr_ah = WR & (ADDR == (BASE_ADDR + 3));
  wire OE_tcntl = RD & (ADDR == (BASE_ADDR + 4));
  wire WE_tcntl = WR & (ADDR == (BASE_ADDR + 4));
  wire OE_tcnth = RD & (ADDR == (BASE_ADDR + 5));
  wire WE_tcnth = WR & (ADDR == (BASE_ADDR + 5));
  wire OE_tccr_b = RD & (ADDR == (BASE_ADDR + 6));
  wire WE_tccr_b = WR & (ADDR == (BASE_ADDR + 6));
  wire OE_tccr_a = RD & (ADDR == (BASE_ADDR + 7));
  wire WE_tccr_a = WR & (ADDR == (BASE_ADDR + 7));

  reg [2:0] cs0;
  reg [2:0] wgm0;
  reg [1:0] com_a;
  reg [1:0] com_b;

`ifdef TIM16_FUNC_MINIMUM
`else
  reg foc_b;
  reg foc_a;
`endif
  wire  [7:0] i_tccr_a = {com_a, com_b, wgm0[2], cs0};
  wire  [7:0] i_tccr_b = {6'b0, wgm0[1:0]};

  reg t_prev;
  wire t_nops = (cs0 == 1);
  wire t_clk =  (cs0 == 2) ? r_prescaler[2] // 1/8
               : (cs0 == 3) ? r_prescaler[5] // 1/64
               : (cs0 == 4) ? r_prescaler[7] // 1/256
               : (cs0 == 5) ? r_prescaler[9] // 1/1024
               : 0;
  wire t_top_ocr_a = ((wgm0 == 2) | (wgm0 == 5) | (wgm0 == 7));
`ifdef TIM16_FUNC_MINIMUM
`else
  wire ocr_chg_top = ((wgm0 == 1) | (wgm0 == 5));
  wire ocr_chg_btm = ((wgm0 == 3) | (wgm0 == 7));
  wire ocr_chg_imm = (wgm0[0] == 0);
`endif
`ifdef OUTPUT_MUX_TRI
  wire [7:0] i_DO = OE_ocr_bl ? r_ocr_bl : OE_ocr_al ? r_ocr_al 
		  : OE_tcntl ? r_tcntl
		  : (OE_ocr_bh | OE_ocr_ah | OE_tcnth) ? r_tmp
		  : OE_tccr_b ? i_tccr_b : OE_tccr_a ? i_tccr_a : 0;

  bufif1(DO,i_DO, (OE_ocr_bl | OE_ocr_al | OE_tcntl 
		 | OE_ocr_bh | OE_ocr_ah | OE_tcnth 
		 | OE_tccr_b | OE_tccr_a ) );
`else
  wire [7:0] i_DO = OE_ocr_bl ? r_ocr_bl : OE_ocr_al ? r_ocr_al
		  : OE_tcntl ? r_tcntl
		  : (OE_ocr_bh | OE_ocr_ah | OE_tcnth) ? r_tmp
		  : OE_tccr_b ? i_tccr_b : OE_tccr_a ? i_tccr_a : 0 ;
  assign DO = i_DO;
`endif

  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	r_ocr_b <= 0; 	 
	r_ocr_a <= 0;
	cs0 <= 0;
	wgm0[2] <= 0;
	com_b <= 0;
	com_a <= 0;
	wgm0[1:0] <= 0;
`ifdef TIM16_FUNC_MINIMUM
`else
	foc_b <= 0;
	foc_a <= 0;
`endif
	r_tmp <= 0;
	r_tcnt <= 0;
	r_tim_ovf   <= 0;
	r_tim_compa <= 0;
	r_tim_compb <= 0;
	t_prev <= 0;
      end
    else
      begin
	if(WE_ocr_bh | WE_ocr_ah | WE_tcnth ) r_tmp <= DI;
	if(WE_ocr_bl) r_ocr_b <= {r_tmp, DI}; 
	if(WE_ocr_al) r_ocr_a <= {r_tmp, DI}; 
	if(WE_tccr_a) 
	begin
	  com_a <= DI[7:6];
	  com_b <= DI[5:4];
	  wgm0[1:0] <= DI[1:0];
	end

	if(WE_tccr_b) 
	begin
	  wgm0[2] <= DI[3];
	  cs0 <= DI[2:0];
`ifdef TIM16_FUNC_MINIMUM
`else
	  foc_a <= DI[7];
	  foc_b <= DI[6];
`endif
	end
`ifdef TIM16_FUNC_MINIMUM
	if(WE_tcntl) r_tcnt <=  {r_tmp, DI}; 
	else
	  begin
	    if (t_nops | (t_clk ^ t_prev) & t_prev)
              begin
		if (t_top_ocr_a & (r_tcnt == r_ocr_a)) r_tcnt <= 16'h00; 
		else 
		   begin
			r_tcnt <= r_tcnt + 16'h01;
 
			r_tim_ovf   <= (r_tcnt + 16'h01 == 16'hffff);
			r_tim_compa <= (r_tcnt + 16'h01 == r_ocr_a);
			r_tim_compb <= (r_tcnt + 16'h01 == r_ocr_b);
		   end
	      end
	  end
`else
NOT IMPLEMENTED YET
`endif

	t_prev <= t_clk;
      end
  end
endmodule

