/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"

// global defines :  RTAVR_INCLUDE_IOR, RTAVR_INCLDUE_RAM, RTAVR_INCLUDE_ROM

`define RTAVR_INCLUDE_GPR
//`define RTAVR_INCLUDE_IOR
//`define RTAVR_INCLUDE_RAM
//`define RTAVR_INCLUDE_ROM
`define RTAVR_INCLUDE_S0
`define RTAVR_INCLUDE_S1

//
// Module Name: rtavr
//   ver 005
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                                        all-in-one  
// Number of Slice Flip Flops:              247
// Number of 4 input LUTs:                  722
// Number of occupied Slices:               472
// Total Number of 4 input LUTs:            762
// Number of bonded IOBs :                   26
//   IOB Flip Flops :                        24
// Number of BUFGMUXs :                       2
// Number used for Dual Port RAMs            16
// Number of RAMB16BWEs :                     2
// Clock to Setup 
//on destination clock CLK2X              13.030

module rtavr (
		input RESET
`ifdef RTAVR_INCLUDE_S0
`else
		, input S0_WB_INST
		, input S0_IDX_INST
		, output S0_VALID
`endif
`ifdef RTAVR_INCLUDE_S1
  `ifdef RTAVR_INCLUDE_INTCONT
  `else
		, input [4:0] INT_VEC
		, input INT_REQ
		, output INT_ACK
  `endif
`else
		, input CMD_LOAD, CMD_LOAD_WR, CMD_STORE
		, input CMD_OP2, CMD_OP1
		, input ALU_WC
		, input [2:0] OP1_SEL
  `ifdef USE_ALU
		, input CMD_ALU
		, input CMD_ALU_WR, CMD_ALU_WRF
		, input CMD_ALU_SUB_ALL
//		, input CMD_ALU_MOV
		, input CMD_ALU_ADD
		, input CMD_ALU_OR
		, input CMD_ALU_EOR
  `else
		, input CMD_OP3
		, input CMD_OP3_WR, CMD_OP3_WRF
		, input CMD_OP2_WR, CMD_OP2_WRF
		, input [2:0] OP3_SEL
		, input [2:0] OP2_SEL
  `endif
		, input CMD_ALU_SUB
		, input CMD_ALU_SBC
		, input CMD_ALU_ADD
		, input CMD_ALU_Z
		, input CMD_ALU_V0
		, input CMD_ALU_NEG
		, input CMD_ALU_OP1

		, input [7:0] IMM_DATA
		, output [15:0] INST

		, input CMD_SBIX;
		, input CMD_SBRX;
		, input CMD_XBI;
		, input CMD_BST;
		, input CMD_BLD;
		, input CMD_SETI;
		, input CMD_CLRI;
		, input XBI_BIT_IN;
		, input [2:0] SBIX_BIT;
		, input [2:0] FLAGS_BIT;
		, output SBIX_BIT_OUT;
		, output FLAGS_BIT_OUT;
  `ifdef RTAVR_INCLUDE_INTCONT
		, output [4:0] INT_VEC
		, output INT_REQ
		, input INT_ACK
  `endif
		, input [11:0] PC
`endif

`ifdef RTAVR_INCLUDE_GPR
		, input CLK2X
  `ifdef USE_DMY_CLOCK
		, output CLK
  `else
		, input CLK
  `endif
`else
		, input CLK
		, output GPR_WE
		, output[7:0] GPR_DI

		, input[7:0] GPR_DOBL
		, input[7:0] GPR_DOBH
		, input[12:0] GPR_JA
		, input[15:0] GPR_INDEX
		, output GPR_PREDEC, GPR_POSTINC
		, output[3:0] GPR_ADDRAL
		, output[3:0] GPR_ADDRAH
		, output[3:0] GPR_ADDRBL
		, output[3:0] GPR_ADDRBH
  `ifdef FLAGS_TUNE_2
		, input GPR_EQ_BL_BH
  `endif
`endif
`ifdef RTAVR_INCLUDE_IOR
		// external I/O
  `ifdef RTAVR_PORT_ESCALATION
		, input [23:0] PIN
		, output [23:0] DDR
		, output [23:0] PORT
  `else
    `ifdef IOR_HAVE_PORTA
		, inout [7:0] PORTA
    `endif
    `ifdef IOR_HAVE_PORTB
		, inout [7:0] PORTB
    `endif
    `ifdef IOR_HAVE_PORTC
		, inout [7:0] PORTC
    `endif
  `endif
  `ifdef IOR_HAVE_INT0
  `else
    `ifdef RTAVR_INCLUDE_INTCONT
		, input IOR_INT0
		, output IOR_RST_INT0
    `else
    `endif
  `endif
`else // RTAVR_INCLUDE_IOR
		, output IOR_WEB
		, output [5:0] IOR_ADDRB
		, output [7:0] IOR_DIB
		, input [7:0] IOR_DOB
		, output IOR_OE
		// SREG(FLAGS) & SP
		, input [7:0] FLAGS_IN
		, output [7:0] FLAGS_OUT
		, input [15:0] SP_IN
		, output [15:0] SP_OUT
  `ifdef RTAVR_INCLUDE_INTCONT
  `else
		, input IOR_INT0
		, output IOR_RST_INT0
    `ifdef IOR_HAVE_INT1
		, input IOR_INT1
		, output IOR_RST_INT1
    `endif
    `ifdef IOR_HAVE_TIM0
		, input IOR_TIM0_OVF
		, input IOR_TIM0_COMPA
		, input IOR_TIM0_COMPB
		, output IOR_RST_TIM0_OVF
		, output IOR_RST_TIM0_COMPA
		, output IOR_RST_TIM0_COMPB
    `endif
    `ifdef IOR_HAVE_SPI
		, input IOR_SPI_INT
		, output IOR_RST_SPI_INT
    `endif
    `ifdef IOR_HAVE_USART
		, input IOR_USART_TX
		, input IOR_USART_UDRE
		, input IOR_USART_RX
		, output IOR_RST_USART_TX
    `endif
    `ifdef IOR_HAVE_TWI
		, input IOR_TWI_SLAVE
		, output IOR_RST_TWI_SLAVE
    `endif
   `endif  // RTAVR_INCLUDE_INTCONT
`endif // RTAVR_INCLUDE_IOR
`ifdef RTAVR_INCLUDE_RAM
`else
		, output RAM_WEB
		, output [10:0] RAM_ADDRB
		, output [7:0] RAM_DIB
		, input [7:0] RAM_DOB
  `ifdef OUTPUT_MUX_NONE
  `else
		, output RAM_OE
  `endif
`endif
`ifdef RTAVR_INCLUDE_ROM
`else
		, output [11:0] ROM_ADDRA
		, input [15:0] ROM_DOA
		, output ROM_WEB
		, output [12:0] ROM_ADDRB

		, output [7:0] ROM_DIB
		, input[7:0] ROM_DOB
  `ifdef OUTPUT_MUX_NONE
  `else
		, output ROM_OE
  `endif
`endif
		, output PW_SLEEP
`ifdef PRC_WATCH_WDR
		, output PW_WDR
`endif
`ifdef PRC_WATCH
		, output PW_STORE
		, output PW_LOAD
		, output [11:0] PW_PC
		, output [7:0] PW_SREG
		, output [11:0] PW_SP
		, output [7:0] PW_LOAD_REG
		, output [3:0] PW_LOAD_ID
		, output [7:0] PW_STORE_MEM
		, output [11:0] PW_STORE_ADDR
`endif
		);

    wire [2:0] SBIX_BIT;
`ifdef RTAVR_INCLUDE_GPR

  wire GPR_WE;
  wire [7:0] GPR_DI;
  wire GPR_PREDEC,GPR_POSTINC;
  wire [4:0] GPR_IDX_OFF;
  wire [3:0] GPR_ADDRAL;
  wire [3:0] GPR_ADDRAH;
  wire [3:0] GPR_ADDRBL;
  wire [3:0] GPR_ADDRBH;

  wire [7:0] GPR_DOBL;
  wire [7:0] GPR_DOBH;
  wire [11:0] GPR_JA;
  wire [15:0] GPR_INDEX;
  `ifdef FLAGS_TUNE_2
	wire GPR_EQ_BL_BH;
  `endif

`ifdef PRC_WATCH
  wire [3:0] WB_ADDR;
`endif

  rtavr_gpr_16 GPR(.RESET(RESET)
`ifdef USE_DMY_CLOCK
		,  .CLK2X_IN(CLK2X)
		,  .CLK_OUT(CLK)
`else
		,  .CLK2X(CLK2X)
		,  .CLK(CLK)
`endif
		, .WE(GPR_WE), .DI(GPR_DI)
                , .PREDEC(GPR_PREDEC), .POSTINC(GPR_POSTINC)
                , .IDX_OFF(GPR_IDX_OFF)
		, .ADDRAL(GPR_ADDRAL[3:0]), .ADDRAH(GPR_ADDRAH[3:0])
		, .ADDRBL(GPR_ADDRBL[3:0]), .ADDRBH(GPR_ADDRBH[3:0])

		, .DOBL(GPR_DOBL[7:0]), .DOBH(GPR_DOBH[7:0])
		, .JA(GPR_JA[11:0])
		, .INDEX(GPR_INDEX[15:0])
  `ifdef FLAGS_TUNE_2
                , .EQ_BL_BH(GPR_EQ_BL_BH)
  `endif
  `ifdef SBRX_TUNE
		, .BL_BIT_SEL(SBIX_BIT[2:0])
		, .BL_BIT_OUT(GPR_BL_BIT_OUT)
  `endif
`ifdef PRC_WATCH
		, .WB_ADDR(WB_ADDR[3:0])
`endif
		); 
`endif
`ifdef IOR_HAVE_RAMWIN
    wire RAM_WEA;
    wire [10:0] RAM_ADDRA;
    wire [7:0] RAM_DIA;
    wire [7:0] RAM_DOA;
`endif
`ifdef RTAVR_INCLUDE_IOR
  wire IOR_WEB;
  wire [5:0] IOR_ADDRB;
  wire [7:0] IOR_DIB;
  wire [7:0] IOR_DOB;

  wire [7:0] FLAGS_IN;
  wire [7:0] FLAGS_OUT;
  wire [15:0] SP_IN;
  wire [15:0] SP_OUT;
  
  `ifdef IOR_HAVE_INT0
    wire IOR_RST_INT0;
    `ifdef IOR_HAVE_INT1
      wire IOR_RST_INT1;
    `endif
  `endif
  `ifdef IOR_HAVE_TIM0
    wire IOR_RST_TIM0_OVF;
    wire IOR_RST_TIM0_COMPA;
    wire IOR_RST_TIM0_COMPB;
  `endif
  `ifdef IOR_HAVE_SPI
    wire IOR_RST_SPI_INT;
  `endif
  `ifdef IOR_HAVE_USART
    wire IOR_RST_USART_TX;
  `endif
  `ifdef IOR_HAVE_TWI
    wire IOR_RST_TWI_SLAVE;
  `endif

  `ifdef RTAVR_PORT_ESCALATION
  `else
     wire [23:0] DDR;
     wire [23:0] PORT;
     wire [23:0] PIN;
  `endif

  rtavr_ior IOR (.CLK(CLK), .RESET(RESET)
  `ifdef IOR_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
  `endif
		, .ADDR(IOR_ADDRB[5:0]), .DI(IOR_DIB[7:0]), .DO(IOR_DOB[7:0])
		, .RD(IOR_OE), .WR(IOR_WEB)
  `ifdef IOR_PORT_ESCALATION
		, .DDR(DDR[23:0])
		, .PORT(PORT[23:0])
		, .PIN(PIN[23:0])
  `else
    `ifdef IOR_HAVE_PORTA
		, .PORTA(PORTA[7:0])
    `endif
    `ifdef IOR_HAVE_PORTB
		, .PORTB(PORTB[7:0])
    `endif
    `ifdef IOR_HAVE_PORTC
		, .PORTC(PORTC[7:0])
    `endif
  `endif
  `ifdef IOR_HAVE_INT0
		, .INT0(IOR_INT0)
		, .RST_INT0(IOR_RST_INT0)
    `ifdef IOR_HAVE_INT1
		, .INT1(IOR_INT1)
		, .RST_INT1(IOR_RST_INT1)
    `endif
  `endif
		// SREG(FLAGS) & SP
		, .FLAGS_IN(FLAGS_OUT[7:0]) , .FLAGS_OUT(FLAGS_IN[7:0])
		, .SP_IN(SP_OUT[15:0]) , .SP_OUT(SP_IN[15:0])
  `ifdef IOR_HAVE_RAMWIN
		, .RAM_WE(RAM_WEA)
		, .RAM_ADDR(RAM_ADDRA)
		, .RAM_DI(RAM_DIA)
		, .RAM_DO(RAM_DOA)
  `endif
  `ifdef IOR_HAVE_TIM0
		// INTERRUPTS 
		, .TIM0_OVF(IOR_TIM0_OVF)
		, .RST_TIM0_OVF(IOR_RST_TIM0_OVF)
		, .TIM0_COMPA(IOR_TIM0_COMPA)
		, .RST_TIM0_COMPA(IOR_RST_TIM0_COMPA)
		, .TIM0_COMPB(IOR_TIM0_COMPB)
		, .RST_TIM0_COMPB(IOR_RST_TIM0_COMPB)
  `endif
  `ifdef IOR_HAVE_SPI
		, .SPI_INT(IOR_SPI_INT)
		, .RST_SPI_INT(IOR_RST_SPI_INT)
  `endif
  `ifdef IOR_HAVE_USART
		, .USART_TX_INT(IOR_USART_TX)
		, .USART_UDRE_INT(IOR_USART_UDRE)
		, .USART_RX_INT(IOR_USART_RX)
		, .RST_USART_TX_INT(IOR_RST_USART_TX)
  `endif
  `ifdef IOR_HAVE_TWI
		, .TWI_SLAVE(IOR_TWI_SLAVE)
		, .RST_TWI_SLAVE(IOR_RST_TWI_SLAVE)
  `endif
  `ifdef SBRI_TUNE
		, .DO_BIT_SEL(SBIX_BIT[2:0])
		, .DO_BIT_OUT(IOR_BIT_OUT)
  `endif
		);
`endif // RTAVR_INCLUDE_IOR
`ifdef RTAVR_INCLUDE_RAM
  wire RAM_WEB;

  wire [10:0] RAM_ADDRB;
  wire [7:0] RAM_DIB;
  wire [7:0] RAM_DOB;

 
  rtavr_sram_2KB RAM(.CLK(CLK), .WEB(RAM_WEB)
  `ifdef RAM_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
  `endif
		, .ADDRB(RAM_ADDRB[10:0])
		, .DIB(RAM_DIB[7:0]), .DOB(RAM_DOB[7:0])
  `ifdef OUTPUT_MUX_NONE
  `else
		, .OE(RAM_OE)
  `endif  
  `ifdef IOR_HAVE_RAMWIN
		, .WEA(RAM_WEA), .ADDRA(RAM_ADDRA[10:0])
		, .DIA(RAM_DIA[7:0]), .DOA(RAM_DOA[7:0])
  `endif
		);
`endif
`ifdef RTAVR_INCLUDE_ROM
  wire ROM_WEB;

  wire [12:0] ROM_ADDRB;
  wire [7:0] ROM_DIB;
  wire [7:0] ROM_DOB;
  
  wire [11:0] ROM_ADDRA;
  wire [15:0] ROM_DOA;
  rtavr_rom #(.SIZE(`ROM_SIZE))
	ROM (.CLK(CLK), .WEB(ROM_WEB)
  `ifdef ROM_DRIVEN_CLK2X
		, .CLK2X(CLK2X)
  `endif
		, .ADDRB(ROM_ADDRB[12:0])
		, .DIB(ROM_DIB[7:0]), .DOB(ROM_DOB[7:0])
		, .ADDRA(ROM_ADDRA[11:0]), .DOA(ROM_DOA[15:0])
  `ifdef OUTPUT_MUX_NONE
  `else
		, .OE(ROM_OE)
  `endif  
		);
`endif

  wire [11:0] PC;
`ifdef RTAVR_INCLUDE_S0
  wire [15:0] INST;
  `ifdef NEXTPC_TUNE_4
      wire [11:0] PD_RADR;
  `elsif NEXTPC_TUNE_3
      wire [11:0] PD_RADR;
  `endif
  wire S0_VALID;
  rtavr_s0_fetch  S0 (.CLK(CLK), .RESET(RESET), .PM_OUT(ROM_DOA[15:0])
                , .S0_VALID(S0_VALID)
                , .WB_INST(S0_WB_INST)
                , .IDX_INST(S0_IDX_INST)
		, .INST(INST[15:0])
                , .GPR_PREDEC(GPR_PREDEC), .GPR_POSTINC(GPR_POSTINC)
                , .GPR_IDX_OFF(GPR_IDX_OFF)
		, .GPR_ADDRAL(GPR_ADDRAL[3:0]), .GPR_ADDRAH(GPR_ADDRAH[3:0])
		, .GPR_ADDRBL(GPR_ADDRBL[3:0]), .GPR_ADDRBH(GPR_ADDRBH[3:0])
`ifdef NEXTPC_TUNE
		, .PD_EXT(PD_EXT)
		, .PD_JUMPS_I(PD_JUMPS_I)
		, .PD_JUMPS_R(PD_JUMPS_R)
		, .PD_SKIPS(PD_SKIPS)
		, .PD_BR(PD_BR)
  `ifdef NEXTPC_TUNE_4
		, .PC(PC[11:0])
		, .PD_RADR(PD_RADR)
  `elsif NEXTPC_TUNE_3
		, .PD_RADR(PD_RADR)
  `endif
  `ifdef EA_TUNE
		, .PD_BSET_BCLR(PD_BSET_BCLR)
		, .PD_EA_ABS(PD_EA_ABS)
		, .PD_PUSH_POP(PD_PUSH_POP)
		, .PD_CALLS(PD_CALLS)
  `endif
`endif
		);
`endif
`ifdef RTAVR_INCLUDE_S1
  `ifdef USE_ALU
  `else
    wire [2:0] OP3_SEL;
    wire [2:0] OP2_SEL;
  `endif
    wire [2:0] OP1_SEL;
    wire [7:0] IMM_DATA;
    wire [7:0] S1_DI;

    wire [2:0] FLAGS_BIT;
    wire [15:0] EA;
  `ifdef RTAVR_INCLUDE_INTCONT
    wire [4:0] INT_VEC;
  `endif

  rtavr_s1_decode S1 (.CLK(CLK), .RESET(RESET)
                , .S0_WB_INST(S0_WB_INST)
                , .S0_IDX_INST(S0_IDX_INST)
                , .S0_VALID(S0_VALID)
                , .INST(INST[15:0])
                , .JA(GPR_JA[11:0])
                , .EA_INDEX(GPR_INDEX[15:0])

                , .CMD_LOAD(CMD_LOAD)
                , .CMD_LOAD_WR(CMD_LOAD_WR)
		, .CMD_STORE(CMD_STORE)
		, .CMD_OP2(CMD_OP2), .CMD_OP1(CMD_OP1)
		, .ALU_WC(ALU_WC)
`ifdef USE_ALU
		, .CMD_ALU(CMD_ALU)
		, .CMD_ALU_WR(CMD_ALU_WR)
		, .CMD_ALU_WRF(CMD_ALU_WRF)
//		, .CMD_ALU_MOV(CMD_ALU_MOV)
		, .CMD_ALU_AND(CMD_ALU_AND)
		, .CMD_ALU_OR(CMD_ALU_OR)
		, .CMD_ALU_EOR(CMD_ALU_EOR)
		, .CMD_ALU_SUB_ALL(CMD_ALU_SUB_ALL)
`else
		, .CMD_OP3(CMD_OP3)
		, .CMD_OP3_WR(CMD_OP3_WR) , .CMD_OP3_WRF(CMD_OP3_WRF)
		, .CMD_OP2_WR(CMD_OP2_WR) , .CMD_OP2_WRF(CMD_OP2_WRF)

		, .OP3_SEL(OP3_SEL[2:0])
		, .OP2_SEL(OP2_SEL[2:0])
`endif
		, .CMD_ALU_SUB(CMD_ALU_SUB)
		, .CMD_ALU_SBC(CMD_ALU_SBC)
		, .CMD_ALU_ADD(CMD_ALU_ADD)
		, .CMD_ALU_Z(CMD_ALU_Z)
		, .CMD_ALU_V0(CMD_ALU_V0)
		, .CMD_ALU_NEG(CMD_ALU_NEG)
		, .CMD_ALU_OP1(CMD_ALU_OP1)

		, .OP1_SEL(OP1_SEL[2:0])
		, .IMM_DATA(IMM_DATA[7:0])
		, .LOAD_DATA(S1_DI)
`ifdef NEXTPC_TUNE
		, .PD_EXT(PD_EXT)
		, .PD_JUMPS_I(PD_JUMPS_I)
		, .PD_JUMPS_R(PD_JUMPS_R)
		, .PD_SKIPS(PD_SKIPS)
		, .PD_BR(PD_BR)
  `ifdef NEXTPC_TUNE_4
		, .PD_RADR(PD_RADR)
  `elsif NEXTPC_TUNE_3
		, .PD_RADR(PD_RADR)
  `endif
  `ifdef EA_TUNE
		, .PD_BSET_BCLR(PD_BSET_BCLR)
		, .PD_EA_ABS(PD_EA_ABS)
		, .PD_PUSH_POP(PD_PUSH_POP)
		, .PD_CALLS(PD_CALLS)
  `endif
`endif
		, .CMD_SBIX(CMD_SBIX)
		, .CMD_SBRX(CMD_SBRX)
		, .CMD_XBI(CMD_XBI)
		, .CMD_BST(CMD_BST)
		, .CMD_BLD(CMD_BLD)
		, .CMD_SETI(CMD_SETI)
		, .CMD_CLRI(CMD_CLRI)
		, .SBIX_BIT(SBIX_BIT[2:0])
		, .SBIX_BIT_IN(SBIX_BIT_OUT)
		, .XBI_BIT_OUT(XBI_BIT_IN)
		, .FLAGS_BIT(FLAGS_BIT[2:0])
		, .FLAGS_BIT_IN(FLAGS_BIT_OUT)

		, .EA(EA[15:0])
		, .SP_OUT(SP_OUT[15:0]), .SP_IN(SP_IN[15:0])

		, .EA_PUSH(EA_PUSH)
		, .PUSH_SEL(PUSH_SEL)

		, .INT_VEC(INT_VEC)
		, .INT_REQ(INT_REQ)
		, .INT_ACK(INT_ACK)

		, .PW_SLEEP(PW_SLEEP)
  `ifdef PRC_WATCH_WDR
		, .PW_WDR(PW_WDR)
  `endif  
		, .PC(PC[11:0])
		);
`endif  
  assign ROM_ADDRA = PC;
  
  // 0x0000 - 0x003f (64B)
  wire oe_ior = (EA[15:6] == 0);

  // 0x0040 - 0x1000 (4KB cyclic)
  wire oe_ram =  (EA[15:12] == 0)  & ~(EA[11:6] == 0);

  // 0x4000 - 0x5fff (8KB)
  wire oe_rom = (EA[15:13] == 2);
    
  wire wr_ior = CMD_STORE & oe_ior;
  wire wr_ram = CMD_STORE & oe_ram;
  wire wr_rom = CMD_STORE & oe_rom;

  wire rd_ior = CMD_LOAD & oe_ior;
  wire rd_ram = CMD_LOAD & oe_ram;
  wire rd_rom = CMD_LOAD & oe_rom;
  
  wire [7:0] Rd_in = GPR_DOBL;
  wire [7:0] Rr_in = GPR_DOBH;

  wire   di_is_zero;
  wire C_out;
  wire Z_out;
  wire N_out;
  wire V_out;
  wire S_out;
  wire H_out;
  wire T_out;
  wire I_out;
  
  wire C_in = FLAGS_IN[0];
  wire Z_in = FLAGS_IN[1];
  wire N_in = FLAGS_IN[2];
  wire V_in = FLAGS_IN[3];
  wire S_in = FLAGS_IN[4];
  wire H_in = FLAGS_IN[5];
  wire T_in = FLAGS_IN[6];
  wire I_in = FLAGS_IN[7];
  reg C_in_l;

always @(negedge CLK)
  begin
    if (RESET == 0)
      begin
	C_in_l <= 0;
      end
    else
      begin
	C_in_l <= C_in;
      end
  end

  assign FLAGS_OUT[0] = C_out;
  assign FLAGS_OUT[1] = Z_out;
  assign FLAGS_OUT[2] = N_out;
  assign FLAGS_OUT[3] = V_out;
  assign FLAGS_OUT[4] = S_out;
  assign FLAGS_OUT[5] = H_out;
  assign FLAGS_OUT[6] = T_out;
  assign FLAGS_OUT[7] = I_out;

// OP3          --*-:**------------                               HSVNZC
// CPSE		000B:00rd:dddd:rrrr	# B=1: CPSE           #   ------
// CP/CPC       000B:01rd:dddd:rrrr	# B=1: CP  B=0: CPC   #   XXXXXX
// SUB/SBC	000B:10rd:dddd:rrrr	# B=1: sub B=0: sbc   #   XXXXXX
// ADD/ADD	000C:11rd:dddd:rrrr	# C=0: add C=1: adc   #   XXXXXX
// AND		0010:00rd:dddd:rrrr	# logical AND         #   -X0XX-
// EOR		0010:01rd:dddd:rrrr	# exclusive OR        #   -X0XX-
// OR		0010:10rd:dddd:rrrr	# inclusive OR        #   -X0XX-
// MOV		0010:11rd:dddd:rrrr	# copy register       #   ------

// # OP2 INST
//              -***:--------------                           #   HSVNZC
// LDI		1110:KKKK:dddd:KKKK	#                     #   ------
// CPI		0011:KKKK:dddd:KKKK	#                     #   XXXXXX 
// SBCI		0100:KKKK:dddd:KKKK	#                     #   XXXXXX 
// SUBI		0101:KKKK:dddd:KKKK	#                     #   XXXXXX
// ORI		0110:KKKK:dddd:KKKK	#                     #   -X0XX- 
// ANDI		0111:KKKK:dddd:KKKK	#                     #   -X0XX-  

`ifdef USE_ALU
  wire [7:0] ALU_RR = CMD_OP2 ? IMM_DATA : Rr_in;
  
      wire [8:0] ALU_OUT = CMD_ALU_ADD ? ({1'b0, Rd_in}  + {1'b0, ALU_RR }
				   + {8'b0, (ALU_WC)? C_in_l : 1'b0 })
		     : CMD_ALU_SUB_ALL ? ({1'b0,Rd_in} - {1'b0, ALU_RR }
				   - {8'b0, (ALU_WC) ? 1'b0: C_in_l })
		     : CMD_ALU_OR ?  {C_in_l, (Rd_in | ALU_RR) }
		     : CMD_ALU_EOR ? {C_in_l, (Rd_in ^ ALU_RR) }
		     : CMD_ALU_AND ? {C_in_l, (Rd_in & ALU_RR) }
		     : {C_in_l, ALU_RR } ;
`else
  wire [8:0] OP3_OUT = (OP3_SEL == 7) ? {C_in_l, Rr_in }
		: (OP3_SEL == 6) ? {C_in_l, (Rd_in | Rr_in) }
		: (OP3_SEL == 5) ? {C_in_l, (Rd_in ^ Rr_in) }
		: (OP3_SEL == 4) ? {C_in_l, (Rd_in & Rr_in) }
		: (OP3_SEL == 3) ? ({1'b0, Rd_in}  + {1'b0, Rr_in}
				   + {8'b0, (ALU_WC)? C_in_l : 1'b0 })
		: ({1'b0,Rd_in} - {1'b0, Rr_in}
				   - {8'b0, (ALU_WC) ? 1'b0: C_in_l }) ;

  wire [8:0] OP2_OUT = (OP2_SEL == 7) ? {C_in_l, (Rd_in & IMM_DATA ) }
		: (OP2_SEL == 6) ? {C_in_l, (Rd_in | IMM_DATA) } 
		: (OP2_SEL == 5) | (OP2_SEL == 4)| (OP2_SEL == 3)
		    ? ({1'b0, Rd_in} - {1'b0, IMM_DATA}
				   - {8'b0, (ALU_WC) ? 1'b0: C_in_l })
		: { C_in_l , IMM_DATA };
`endif

// OP1          --------------:0***                           #   HSVNZC
// COM		1001:010d:dddd:0000	#                     #   -X0XX1
// NEG		1001:010d:dddd:0001	#                     #   XXXXXX 
// SWAP		1001:010d:dddd:0010	#                     #   ------
// INC		1001:010d:dddd:0011	#                     #   -XXXX-
// DEC		1001:010d:dddd:1010	#  (1010 -> 0100 )    #   -XXXX-
// ASR		1001:010d:dddd:0101	#                     #   -XXXXX
// LSR		1001:010d:dddd:0110	#                     #   -XX0XX
// ROR		1001:010d:dddd:0111	#                     #   -XXXXX

  wire [8:0] OP1_OUT = (OP1_SEL == 7) ? { Rd_in[0], C_in_l , Rd_in[7:1] }
		: (OP1_SEL == 6) ? { Rd_in[0] , 1'b0, Rd_in[7:1] }
		: (OP1_SEL == 5) ? { Rd_in[0] , Rd_in[7] , Rd_in[7:1] }
		: (OP1_SEL == 4) ? { C_in_l , Rd_in - 8'h01 }
		: (OP1_SEL == 3) ? { C_in_l , Rd_in + 8'h01 }
		: (OP1_SEL == 2) ? { 1'b0 , Rd_in[3:0] , Rd_in[7:4] }
		: (OP1_SEL == 1) ? 0 -  { 1'b0 , Rd_in }
		: (OP1_SEL == 0) ? { 1'b1 , ~Rd_in } : 0;
 
  wire op1_v   =  (OP1_SEL == 3) ? ( Rd_in == 8'b01111111 )  // INC
                : (OP1_SEL == 1) | (OP1_SEL == 4) ?          // NEG, DEC
                                   ( Rd_in == 8'b10000000 )
		: (OP1_SEL == 2) ? V_in                      // SWAP
		: (OP1_SEL == 0) ? 1'b0                      // COM
                : OP1_OUT[7] ^ Rd_in[0];                // ASR , LSR , ROR

`ifdef SBRX_TUNE
  wire v_sbrx_bit = GPR_BL_BIT_OUT;
  wire [7:0] v_bld_out; 
  assign v_bld_out[0] = (SBIX_BIT == 0) ? T_in : GPR_BL_BIT_OUT;
  assign v_bld_out[1] = (SBIX_BIT == 1) ? T_in : GPR_BL_BIT_OUT;
  assign v_bld_out[2] = (SBIX_BIT == 2) ? T_in : GPR_BL_BIT_OUT;
  assign v_bld_out[3] = (SBIX_BIT == 3) ? T_in : GPR_BL_BIT_OUT;
  assign v_bld_out[4] = (SBIX_BIT == 4) ? T_in : GPR_BL_BIT_OUT;
  assign v_bld_out[5] = (SBIX_BIT == 5) ? T_in : GPR_BL_BIT_OUT;
  assign v_bld_out[6] = (SBIX_BIT == 6) ? T_in : GPR_BL_BIT_OUT;
  assign v_bld_out[7] = (SBIX_BIT == 7) ? T_in : GPR_BL_BIT_OUT;
`else
  wire v_sbrx_bit =  (SBIX_BIT == 0) ? Rd_in[0]
                   : (SBIX_BIT == 1) ? Rd_in[1] 
		   : (SBIX_BIT == 2) ? Rd_in[2]
		   : (SBIX_BIT == 3) ? Rd_in[3]
		   : (SBIX_BIT == 4) ? Rd_in[4]
		   : (SBIX_BIT == 5) ? Rd_in[5]
		   : (SBIX_BIT == 6) ? Rd_in[6]
		   : Rd_in[7];

  wire [7:0] v_bld_out; 
  assign v_bld_out[0] = (SBIX_BIT == 0) ? T_in : Rd_in[0];
  assign v_bld_out[1] = (SBIX_BIT == 1) ? T_in : Rd_in[1];
  assign v_bld_out[2] = (SBIX_BIT == 2) ? T_in : Rd_in[2];
  assign v_bld_out[3] = (SBIX_BIT == 3) ? T_in : Rd_in[3];
  assign v_bld_out[4] = (SBIX_BIT == 4) ? T_in : Rd_in[4];
  assign v_bld_out[5] = (SBIX_BIT == 5) ? T_in : Rd_in[5];
  assign v_bld_out[6] = (SBIX_BIT == 6) ? T_in : Rd_in[6];
  assign v_bld_out[7] = (SBIX_BIT == 7) ? T_in : Rd_in[7];
`endif

`ifdef SBRI_TUNE
  wire v_sbri_bit =  IOR_BIT_OUT;
`else
  wire v_sbri_bit =  (SBIX_BIT == 0) ? IOR_DOB[0]
                   : (SBIX_BIT == 1) ? IOR_DOB[1] 
		   : (SBIX_BIT == 2) ? IOR_DOB[2]
		   : (SBIX_BIT == 3) ? IOR_DOB[3]
		   : (SBIX_BIT == 4) ? IOR_DOB[4]
		   : (SBIX_BIT == 5) ? IOR_DOB[5]
		   : (SBIX_BIT == 6) ? IOR_DOB[6]
		   : IOR_DOB[7];
`endif

  assign SBIX_BIT_OUT = CMD_SBIX ? v_sbri_bit 
                       : CMD_SBRX ? v_sbrx_bit
  `ifdef FLAGS_TUNE_2
		       : GPR_EQ_BL_BH;
  `elsif FLAGS_TUNE
		       : (Rd_in == Rr_in) ;
  `else
		       : di_is_zero ;
  `endif

  assign IOR_OE = rd_ior;
`ifdef OUTPUT_MUX_NONE
`else
  assign RAM_OE = rd_ram;
  assign ROM_OE = rd_rom;
`endif

  assign IOR_WEB = wr_ior;
  assign RAM_WEB = wr_ram;
  assign ROM_WEB = wr_rom;

  assign IOR_ADDRB = EA[5:0];
  assign RAM_ADDRB = EA[10:0];
  assign ROM_ADDRB = EA[11:0];

  wire [7:0] v_ior_dob2;
  assign v_ior_dob2[0] = (SBIX_BIT == 0) ? XBI_BIT_IN : IOR_DOB[0];
  assign v_ior_dob2[1] = (SBIX_BIT == 1) ? XBI_BIT_IN : IOR_DOB[1];
  assign v_ior_dob2[2] = (SBIX_BIT == 2) ? XBI_BIT_IN : IOR_DOB[2];
  assign v_ior_dob2[3] = (SBIX_BIT == 3) ? XBI_BIT_IN : IOR_DOB[3];
  assign v_ior_dob2[4] = (SBIX_BIT == 4) ? XBI_BIT_IN : IOR_DOB[4];
  assign v_ior_dob2[5] = (SBIX_BIT == 5) ? XBI_BIT_IN : IOR_DOB[5];
  assign v_ior_dob2[6] = (SBIX_BIT == 6) ? XBI_BIT_IN : IOR_DOB[6];
  assign v_ior_dob2[7] = (SBIX_BIT == 7) ? XBI_BIT_IN : IOR_DOB[7];

   `ifdef OUTPUT_MUX_NONE
         wire [7:0] load_data =  CMD_XBI ? v_ior_dob2 
			: rd_ior ? IOR_DOB : rd_ram ? RAM_DOB 
			: rd_rom ? ROM_DOB : 0;
   `else
     `ifdef OUTPUT_MUX_OR
         wire [7:0] load_data =  CMD_XBI ? v_ior_dob2 
			: rd_ior ? IOR_DOB 
			: RAM_DOB | ROM_DOB;
      `else
         wire [7:0] load_data =  CMD_XBI ? v_ior_dob2 
			: rd_ior ? IOR_DOB : rd_ram ? RAM_DOB 
			: rd_rom ? ROM_DOB : 0;
      `endif
   `endif

  wire [7:0] store_data = ( EA_PUSH & PUSH_SEL) ? IMM_DATA 
			: (CMD_LOAD & CMD_STORE) ? v_ior_dob2
			: Rd_in;
  
  assign IOR_DIB = store_data;
  assign RAM_DIB = store_data;
  assign ROM_DIB = store_data;
  
  assign S1_DI = load_data;
  assign GPR_DI =  CMD_LOAD ? load_data
`ifdef USE_ALU
		 : CMD_ALU ? ALU_OUT[7:0] 
`else
		 : CMD_OP3 ? OP3_OUT[7:0] 
                 : CMD_OP2 ? OP2_OUT[7:0]
`endif
		 : CMD_OP1 ? OP1_OUT[7:0]
                 : CMD_BLD ? v_bld_out[7:0] : 8'h00 ;

`ifdef USE_ALU
  assign GPR_WE = CMD_LOAD_WR | CMD_ALU_WR | CMD_OP1 | CMD_BLD ;
`else
  assign GPR_WE = CMD_LOAD_WR | CMD_OP3_WR | CMD_OP2_WR | CMD_OP1 | CMD_BLD ;
`endif
 
`ifdef USE_ALU
  wire Rr_in_3  = ALU_RR[3];
  wire Rr_in_7  = ALU_RR[7];

  `ifdef FLAGS_TUNE
	wire [7:0] Rd_out = ALU_OUT[7:0];  // +12 slices
	assign di_is_zero = 
		 CMD_OP1 ? ( OP1_OUT[7:0] == 0)
		 : ( ALU_OUT[7:0] == 0) ;
  `else
	wire [7:0] Rd_out = GPR_DI[7:0];
	assign di_is_zero = (GPR_DI == 0);
  `endif

      assign C_out = CMD_ALU_WRF & ALU_OUT[8] | CMD_OP1 & OP1_OUT[8] 
		 | ~(CMD_ALU_WRF|CMD_OP1) & C_in_l ; 

      assign Z_out = ( CMD_ALU_SBC & Z_in | CMD_ALU_Z ) & di_is_zero 
  		 | ~(CMD_ALU_SBC | CMD_ALU_Z ) & Z_in ;

      assign N_out = (CMD_ALU_WRF | CMD_OP1) & GPR_DI[7]
		  | ~(CMD_ALU_WRF | CMD_OP1) & N_in;
      assign S_out = (CMD_ALU_WRF | CMD_OP1) & (N_out ^ V_out)
		  | ~(CMD_ALU_WRF | CMD_OP1) & S_in;

      assign V_out = CMD_ALU_ADD & (Rd_out[7] & ~Rd_in[7] & ~Rr_in_7
                		| ~Rd_out[7] &  Rd_in[7] &  Rr_in_7 )
	      | CMD_ALU_SUB  & (   Rd_out[7] & ~Rd_in[7] &  Rr_in_7
               			| ~Rd_out[7] &  Rd_in[7] & ~Rr_in_7 )
	      | CMD_ALU_OP1 & op1_v
	      | ~(CMD_ALU_ADD | CMD_ALU_SUB | CMD_ALU_V0 | CMD_ALU_OP1) & V_in;

      assign H_out = CMD_ALU_ADD & ( 
			  ~Rd_out[3] & Rd_in[3] 
			| ~Rd_out[3] & Rr_in_3
			|  Rd_in[3]  & Rr_in_3 )
	       | CMD_ALU_SUB & (
			   Rd_out[3] & ~Rd_in[3] 
			|  Rd_out[3] &  Rr_in_3
			| ~Rd_in[3]  &  Rr_in_3 )
  `ifdef FLAGS_TUNE
	       | CMD_ALU_NEG & ( OP1_OUT[3] | Rd_in[3] )
  `else
	       | CMD_ALU_NEG & ( GPR_DI[3] | Rd_in[3] )
  `endif
	       | ~(CMD_ALU_ADD | CMD_ALU_SUB | CMD_ALU_NEG) & H_in;

  assign T_out = CMD_BST & v_sbrx_bit | ~CMD_BST & T_in;
  assign I_out = CMD_SETI & 1'b1 | ~(CMD_SETI | CMD_CLRI) & I_in;
`else // USE_ALU
  wire Rr_in_3  = CMD_OP3_WRF  & Rr_in[3] | CMD_OP2_WRF  & IMM_DATA[3];
  wire Rr_in_7  = CMD_OP3_WRF  & Rr_in[7] | CMD_OP2_WRF  & IMM_DATA[7];

  assign C_out = CMD_OP3_WRF & OP3_OUT[8] | CMD_OP2_WRF & OP2_OUT[8] 
                 | CMD_OP1 & OP1_OUT[8] 
		 | ~(CMD_OP3_WRF|CMD_OP2_WRF|CMD_OP1)  & C_in_l ; 

  assign Z_out = ( CMD_ALU_SBC & Z_in | CMD_ALU_Z ) & di_is_zero 
  		 | ~(CMD_ALU_SBC | CMD_ALU_Z ) & Z_in ;

  assign N_out = (CMD_OP3_WRF | CMD_OP2_WRF | CMD_OP1) & GPR_DI[7]
		  | ~(CMD_OP3_WRF | CMD_OP2_WRF | CMD_OP1) & N_in;
  assign S_out = (CMD_OP3_WRF | CMD_OP2_WRF | CMD_OP1) & (N_out ^ V_out)
		  | ~(CMD_OP3_WRF | CMD_OP2_WRF | CMD_OP1) & S_in;

  assign V_out = CMD_ALU_ADD & (   GPR_DI[7] & ~Rd_in[7] & ~Rr_in_7
                		| ~GPR_DI[7] &  Rd_in[7] &  Rr_in_7 )
	      | CMD_ALU_SUB  & (   GPR_DI[7] & ~Rd_in[7] &  Rr_in_7
               			| ~GPR_DI[7] &  Rd_in[7] & ~Rr_in_7 )
	      | CMD_ALU_OP1 & op1_v
	      | ~(CMD_ALU_ADD | CMD_ALU_SUB | CMD_ALU_V0 | CMD_ALU_OP1) & V_in;

  assign H_out = CMD_ALU_ADD & ( 
			  ~GPR_DI[3] & Rd_in[3] 
			| ~GPR_DI[3] & Rr_in_3
			|  Rd_in[3]  & Rr_in_3 )
	       | CMD_ALU_SUB & (
			   GPR_DI[3] & ~Rd_in[3] 
			|  GPR_DI[3] &  Rr_in_3
			| ~Rd_in[3]  &  Rr_in_3 )
	       | CMD_ALU_NEG & ( GPR_DI[3] | Rd_in[3] )
//	       | CMD_ALU_NEG & ( OP1_PUT[3] | Rd_in[3] )
	       | ~(CMD_ALU_ADD | CMD_ALU_SUB | CMD_ALU_NEG) & H_in;

  assign T_out = CMD_BST & v_sbrx_bit | ~CMD_BST & T_in;
  assign I_out = CMD_SETI & 1'b1 | ~(CMD_SETI | CMD_CLRI) & I_in;

`endif // USE_ALU

  wire v_flags_bit = (FLAGS_BIT == 0) ? C_out
		   : (FLAGS_BIT == 1) ? Z_out
		   : (FLAGS_BIT == 2) ? N_out
		   : (FLAGS_BIT == 3) ? V_out
		   : (FLAGS_BIT == 4) ? S_out
		   : (FLAGS_BIT == 5) ? H_out
		   : (FLAGS_BIT == 6) ? T_out
		   : I_in ;

  assign FLAGS_BIT_OUT =  v_flags_bit;

`ifdef PRC_WATCH
   assign PW_STORE = RAM_WEB;
   assign PW_LOAD = GPR_WE;
   assign PW_PC[11:0] = PC[11:0];
   assign PW_SREG[7:0] = FLAGS_OUT[7:0];
   assign PW_SP[11:0] = SP_OUT[11:0];
   assign PW_LOAD_REG[7:0] = GPR_DI[7:0];
   assign PW_LOAD_ID[3:0] = WB_ADDR[3:0];
   assign PW_STORE_MEM[7:0] = RAM_DIB[7:0]; 
//   assign PW_STORE_ADDR[11:0] = {1'b0, RAM_ADDRB[10:0] } ;
   assign PW_STORE_ADDR[11:0] = EA[11:0];
`endif

//`define INT0_vect_num  	 1
//`define INT1_vect_num  	 2
//`define INT2_vect_num  	 3
//`define INT3_vect_num  	 4
//`define TIM0_COMPA_vect_num  	11
//`define TIM0_COMPB_vect_num  	12
//`define TIM0_OVF_vect_num  	13
//`define USART_TX_vect_num     14
//`define USART_UDRE_vect_num   15
//`define TWI_SLAVE_vect_num	16
//`define SPI_vect_num  	17
//`define USART_RX_vect_num     18

`ifdef RTAVR_INCLUDE_INTCONT
  reg [4:0] r_int_vec;  

  wire [4:0] v_int_vec =   IOR_INT0 ?  `INT0_vect_num
`ifdef IOR_HAVE_INT1
                  :        IOR_INT1 ?  `INT1_vect_num
`endif
`ifdef IOR_HAVE_TIM0
                  :  IOR_TIM0_COMPA ? `TIM0_COMPA_vect_num
                  :  IOR_TIM0_COMPB ? `TIM0_COMPB_vect_num
                  :  IOR_TIM0_OVF   ? `TIM0_OVF_vect_num
`endif
`ifdef IOR_HAVE_USART
                  :  IOR_USART_TX   ? `USART_TX_vect_num
                  :  IOR_USART_UDRE ? `USART_UDRE_vect_num
`endif
`ifdef IOR_HAVE_TWI
                  :  IOR_TWI_SLAVE  ? `TWI_SLAVE_vect_num
`endif
`ifdef IOR_HAVE_SPI
                  :  IOR_SPI_INT    ?  `SPI_vect_num
`endif
`ifdef IOR_HAVE_USART
                  :  IOR_USART_RX   ? `USART_RX_vect_num
`endif
		  : 0;
  assign INT_REQ = I_out & (IOR_INT0 
`ifdef IOR_HAVE_INT1
			   | IOR_INT1
`endif
`ifdef IOR_HAVE_TIM0
                           | IOR_TIM0_OVF | IOR_TIM0_COMPA | IOR_TIM0_COMPB
`endif
`ifdef IOR_HAVE_SPI
			   | IOR_SPI_INT
`endif
`ifdef IOR_HAVE_USART
			   | IOR_RST_USART_TX
`endif
`ifdef IOR_HAVE_TWI
			   | IOR_TWI_SLAVE
`endif
			   );
  assign INT_VEC = r_int_vec;

  assign IOR_RST_INT0        = INT_ACK & (r_int_vec == `INT0_vect_num);
`ifdef IOR_HAVE_INT1
  assign IOR_RST_INT1        = INT_ACK & (r_int_vec == `INT1_vect_num);
`endif
`ifdef IOR_HAVE_TIM0
  assign IOR_RST_TIM0_COMPA  = INT_ACK & (r_int_vec == `TIM0_COMPA_vect_num);
  assign IOR_RST_TIM0_COMPB  = INT_ACK & (r_int_vec == `TIM0_COMPB_vect_num);
  assign IOR_RST_TIM0_OVF    = INT_ACK & (r_int_vec == `TIM0_OVF_vect_num);
`endif
`ifdef IOR_HAVE_USART
  assign IOR_RST_USART_TX    = INT_ACK & (r_int_vec == `USART_TX_vect_num);
`endif
`ifdef IOR_HAVE_TWI
  assign IOR_RST_TWI_SLAVE   = INT_ACK & (r_int_vec == `TWI_SLAVE_vect_num);
`endif
`ifdef IOR_HAVE_SPI
  assign IOR_RST_SPI_INT     = INT_ACK & (r_int_vec == `SPI_vect_num);
`endif

always @(negedge CLK)
  begin
    if (RESET == 0)
      begin
	r_int_vec <= 0;
      end
    else
      begin
	if (~INT_ACK) r_int_vec <= v_int_vec;
      end
  end
`endif // RTAVR_INCLUDE_INTCONT


`ifdef RTAVR_PORT_ESCALATION
`elsif IOR_PORT_ESCALATION
  generate 
    genvar i;
  `ifdef IOR_HAVE_PORTA
    for (i=0; i<8; i=i+1)
       begin : port_a_out
	assign PORTA[i] = DDR[i] ? PORT[i]
		       : (~DDR[i] & PORT[i]) ? "H"
		       : "Z";
	assign PIN[i] = DDR[i] ? PORT[i] : PORTA[i];
       end
  `endif
  `ifdef IOR_HAVE_PORTB
    for (i=8; i<16; i=i+1)
       begin : port_b_out
	assign PORTB[i-8] = DDR[i] ? PORT[i]
		       : (~DDR[i] & PORT[i]) ? "H"
		       : "Z";
	assign PIN[i] = DDR[i] ? PORT[i] : PORTB[i-8];
       end
  `endif
  `ifdef IOR_HAVE_PORTC
    for (i=16; i<24; i=i+1)
       begin : port_c_out
	assign PORTC[i-16] = DDR[i] ? PORT[i]
		       : (~DDR[i] & PORT[i]) ? "H"
		       : "Z";
	assign PIN[i] = DDR[i] ? PORT[i] : PORTC[i-16];
       end
  `endif
  endgenerate
`endif

endmodule
