/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : USE_DMY_CLOCK, PRC_WATCH, GPR_INITCLR

//`define USE_DMY_CLOCK
//`define PRC_WATCH
//`define GPR_INITCLR

// local defines
//`define GPR_TUNE_1
//        -- marge  2 ports into one : need define STALL_WB_CONFLICT_2
//        -- not recomended
//	gpr_lo[r_addral] 
//	gpr_lo[r_wb_addr] 


// GPR modele for the Reduced Core TinyAVR (attiny10) 
// version 005
// 
// Target Device: xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal: Balanced
// LATCH_DOL : Insert LATCHes for Timing
//                                 
// Number of Slice Flip Flops:             
// Number of 4 input LUTs:                
// Number of occupied Slices:            
// Total Number of 4 input LUTs:        
// Number used for Dual Port RAMs:     
// Number of bonded IOBs:             
// Post-PAR Static Timing Report
// Clock to Setup (CLK2X) MAX  
//
//
module rtavr_gpr_16(
		input RESET,
`ifdef USE_DMY_CLOCK
		input CLK2X_IN,
		output CLK_OUT,
`else
		input CLK2X,
		input CLK,
`endif
		input WE,
		input PREDEC, POSTINC, 
		input [4:0] IDX_OFF,
		input [7:0] DI,
		input [3:0] ADDRAL,
		input [3:0] ADDRAH,
		input [3:0] ADDRBL,
		input [3:0] ADDRBH,
		output [7:0] DOBL,
		output [7:0] DOBH
		, output [11:0] JA
		, output [15:0] INDEX
  `ifdef FLAGS_TUNE_2
		, output EQ_BL_BH
  `endif
  `ifdef SBRX_TUNE
		, input[2:0] BL_BIT_SEL
		, output BL_BIT_OUT
  `endif
  
`ifdef PRC_WATCH
		, output [3:0] WB_ADDR
`endif
`ifdef S1_SKIP_GPR
		, input [2:0] RD_BIT
		, output RD_BIT_OUT
		, output MATCH_RDRR
`endif
		);
// Timing MEMO:
//
//                T(N-1)               |            TN                 |
//              _______         _______         _______         _______
// CLK2X ______|       |_______|       |_______|       |_______|       |
//                       _______________                 _______________
// CLK _________________|               |_______________|               |
//
// ADDRAL,      ================================ (PM_DRIVEN_CLK2X) 
// ADDRBL                ================================ (else)     
// ADDRAH,ADDRBH,                       =================================
// PREDEC,POSTINC,WE
// DI                                                   ================
//                                     V                       V
//                                 ADDRAL                     DI
//                                 ADDRBL                  
//                   
// DOBL,DOBL                             
// INDEX(lo8)                           
// INDEX(hi8)                          
// JA(lo8)                            
// JA(hi8)                           
//
// READ OPs
//  
//  WE  PREDEC POSTiNC     DOBH(= Rr_in)      DOBL(= Rd_in)
//  0     -      -         gpr[ADDRBH]        gpr[ADDRBL] 
//  1     -      -         gpr[ADDRBH]/DI     gpr[ADDRBL]/DI 
//  
//  JA : ( non-latched { DOBH , DOBL }
//  
//  WE  PREDEC POSTINC     INDEX
//  0     0      0       { gpr[ADDRBH]    ,    gpr[ADDRAL]    }
//  0     1      0       { gpr[ADDRAH]    ,    gpr[ADDRAL]    } -1
//  0     0      1       { gpr[ADDRAH]    ,    gpr[ADDRAL]    }
//  
//  1     0      0       { gpr[ADDRBH]/DI ,    gpr[ADDRAL]/DI }
//  
// WRITE OPs
//  WE  PREDEC POSTINC     
//  0     1      0     --{ gpr[ADDRAH]    ,    gpr[ADDRAL]    } 
//  0     0      1       { gpr[ADDRAH]    ,    gpr[ADDRAL]    } ++
//  
//  1     0      0       gpr[prev ADDRAH] = DI
//  

`ifdef USE_DMY_CLOCK
  reg CLK;
  wire CLK2X;
  assign CLK_OUT = CLK;
  assign CLK2X   = CLK2X_IN;
`endif
  reg  [2:0] r_addral;
  reg  r_addrbl_0;
  reg  [2:0] r_addrbl;
  reg  [2:0] r_wb_addr;
  reg  r_wb_addr_0;
  reg  [2:0] s2_addrbl;
  reg  s2_addrbl_0;
  reg  [2:0] s2_addrbh;
  reg  s2_addrbh_0;

  reg  [7:0] s2_DOBL;
  reg  [7:0] s2_DOBH;
  reg  [7:0] gpr_lo [0:7];
  reg  [7:0] gpr_hi [0:7];
  reg  [15:0] r_doa2;

  assign DOBL = s2_DOBL;
  assign DOBH = s2_DOBH;

  wire WB = (PREDEC | POSTINC);

    wire [15:0] DOA = {  gpr_hi[r_wb_addr] ,  gpr_lo[r_addral] };


  `ifdef PRC_WATCH
    assign WB_ADDR = { r_wb_addr ,r_wb_addr_0 };
  `endif

  wire [15:0] DOA2 = (PREDEC) ? DOA - 1 : DOA;
  wire [15:0] WB_DATA = (PREDEC) ?  DOA - 1
                     : (POSTINC) ?  DOA + 1
		     : DOA;

  wire WE_AL = WB | (WE & ~r_wb_addr_0);
  wire WE_AH = WB | (WE &  r_wb_addr_0);

  wire [7:0] DI2_AL = (WB) ? WB_DATA[7:0] : DI;
  wire [7:0] DI2_AH = (WB) ? WB_DATA[15:8] : DI;

  wire [7:0] v_DOBL = s2_addrbl_0 ? gpr_hi[s2_addrbl] : gpr_lo[s2_addrbl];
  wire [7:0] v_DOBH = s2_addrbh_0 ? gpr_hi[s2_addrbh] : gpr_lo[s2_addrbh];

  assign JA[7:0] = gpr_lo[s2_addrbl];
  assign JA[11:8] = gpr_hi[s2_addrbh];

  assign INDEX[15:13] = (WB) ? r_doa2[15:13] : gpr_hi[s2_addrbh][7:5];
  assign INDEX[12:0] = { (WB) ? r_doa2[12:8] : gpr_hi[s2_addrbh][4:0] 
			, r_doa2[7:0] }
		     + { 8'b00000000, IDX_OFF };

  `ifdef FLAGS_TUNE_2
	reg s2_eq_bl_bh;
	assign EQ_BL_BH = s2_eq_bl_bh;
  `endif
  `ifdef SBRX_TUNE
	reg s2_bl_bit_out;
	assign BL_BIT_OUT = s2_bl_bit_out;

	wire v_bl_bit_out =  (BL_BIT_SEL == 0) ? v_DOBL[0]
			   : (BL_BIT_SEL == 1) ? v_DOBL[1] 
			   : (BL_BIT_SEL == 2) ? v_DOBL[2]
			   : (BL_BIT_SEL == 3) ? v_DOBL[3]
			   : (BL_BIT_SEL == 4) ? v_DOBL[4]
			   : (BL_BIT_SEL == 5) ? v_DOBL[5]
			   : (BL_BIT_SEL == 6) ? v_DOBL[6]
			   : v_DOBL[7];

  `endif
`ifdef USE_DMY_CLOCK
always @(negedge CLK2X)
  begin
	CLK <= ~CLK;
  end

initial
  begin
	CLK <= 0;
  end
`endif

always @(posedge CLK2X)
  begin
    if (RESET == 0)
      begin
	r_wb_addr <= 0;
	r_wb_addr_0 <= 0;
	r_addral <= 0;
	s2_addrbl <= 0;
	s2_addrbl_0 <= 0;
	s2_addrbh <= 0;
	s2_addrbh_0 <= 0;
	s2_DOBL <= 0;
	s2_DOBH <= 0;
  `ifdef FLAGS_TUNE_2
	s2_eq_bl_bh <= 0;
  `endif
  `ifdef SBRX_TUNE
	s2_bl_bit_out <= 0;
  `endif
      end
    else
      begin
	if (~CLK) r_wb_addr_0 <=  (WE)? ADDRAH[0] : ADDRBH[0];
	if (~CLK) r_wb_addr   <=  (WE)? ADDRAH[3:1] : ADDRBH[3:1];
	if (CLK & WE_AH) gpr_hi[r_wb_addr] <= DI2_AH;
`ifdef GPR_TUNE_1
	if (~CLK) r_addral  <=  (~ADDRAH[0] & WE) ? ADDRAH[3:1] : ADDRAL[3:1];
	if (CLK & WE_AL) gpr_lo[r_addral] <= DI2_AL;
`else
	if (~CLK) r_addral    <=  ADDRAL[3:1];
	if (CLK & WE_AL) gpr_lo[r_wb_addr] <= DI2_AL;
`endif
	if (CLK) s2_addrbl_0 <= r_addrbl_0;
	if (CLK) s2_addrbl   <= r_addrbl;
	if (CLK) s2_addrbh_0 <= ADDRBH[0];
	if (CLK) s2_addrbh   <= ADDRBH[3:1];
	s2_DOBL <= v_DOBL;
	s2_DOBH <= v_DOBH;
  `ifdef FLAGS_TUNE_2
	s2_eq_bl_bh <= ( v_DOBL == v_DOBH );
  `endif
  `ifdef SBRX_TUNE
	s2_bl_bit_out <= v_bl_bit_out;
  `endif
      end
  end

always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	r_doa2 <= 0;
      end
    else
      begin
	r_doa2 <= DOA2;
      end
  end

always @(negedge CLK)
  begin
    if (RESET == 0)
      begin
	r_addrbl_0 <= 0;
	r_addrbl   <= 0;
      end
    else
      begin
	r_addrbl_0 <= ADDRBL[0];
	r_addrbl   <= ADDRBL[3:1];
      end
  end

`ifdef GPR_INITCLR
  integer i;
  initial
    begin
      for (i = 0 ; i < 8 ; i = i + 1)
	begin
	  gpr_lo[i] <= 8'h00;
	  gpr_hi[i] <= 8'h00;
	end
    end
`endif
endmodule
