/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_TRI, IOR_HAVE_PORTA , IOR_HAVE_PORTB
//                 , IOR_HAVE_PORTC, IOR_HAVE_INT0, IOR_HAVE_SREG_SP

`define IOR_DO_SYNC
//`define OUTPUT_MUX_TRI
//`define IOR_HAVE_PORTA
//`define IOR_HAVE_PORTB
//`define IOR_HAVE_PORTC
//`define IOR_HAVE_SPI
//`define IOR_HAVE_USART
//`define IOR_HAVE_TWI

//`define IOR_HAVE_INT0
//`define IOR_HAVE_TIM0
//`define IOR_HAVE_TIM0_OC0A
//`define IOR_HAVE_TIM0_OC0B
//`define IOR_HAVE_SREG_SP

// Module Name: rtavr_ior
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//                              NO_PORT  +PORTA +PORTB +PORTC +SREG/SP
// Number of Slice Flip Flops:     61      77     93     112     136
// Number of 4 input LUTs:         97     125    143     176     223
// Number of occupied Slices:      81     104    122     152     175
// Total Number of 4 input LUTs:  106     134    152     185     232
// Number of bonded IOBs :         34      42     50      58     106
//   IOB Flip Flops :               -      16     32      47      71
// Number of BUFGMUXs :             1       1      1       1       1
// Number of RAMB16BWEs :           -       -      -       -       -
// CLK to Setup                 5.921                          5.962
// PORTC -- PORTC + INT0

module rtavr_ior (
		input CLK,
		input RESET,
`ifdef IOR_DRIVEN_CLK2X
		input CLK2X,
`endif
		input WR,RD, 
		input [7:0] DI,
		input [5:0] ADDR,
		output [7:0] DO

`ifdef IOR_HAVE_SREG_SP
		, input  [7:0] FLAGS_IN
		, input  [15:0] SP_IN
		, output [7:0] FLAGS_OUT
		, output [15:0] SP_OUT
`endif
		// external I/O
`ifdef IOR_PORT_ESCALATION
		, output [23:0] DDR
		, output [23:0] PORT
		, input  [23:0] PIN
`else
  `ifdef IOR_HAVE_PORTA
		, inout [7:0] PORTA
  `endif
  `ifdef IOR_HAVE_PORTB
		, inout [7:0] PORTB
  `endif
  `ifdef IOR_HAVE_PORTC
		, inout [7:0] PORTC
  `endif
`endif
		// interrupt handling
`ifdef IOR_HAVE_INT0
		, input  RST_INT0
		, output INT0
  `ifdef IOR_HAVE_INT1
		, input  RST_INT1
		, output INT1
  `endif
`endif
`ifdef IOR_HAVE_RAMWIN
		, output RAM_WE
		, output [10:0] RAM_ADDR
		, output [7:0] RAM_DI
		, input [7:0] RAM_DO
`endif

`ifdef IOR_HAVE_TIM0
		, input  RST_TIM0_OVF
		, output TIM0_OVF
		, input  RST_TIM0_COMPA	
		, output TIM0_COMPA	
		, input  RST_TIM0_COMPB
		, output TIM0_COMPB
`endif
`ifdef IOR_HAVE_SPI
		, input  RST_SPI_INT
		, output SPI_INT
`endif
`ifdef IOR_HAVE_USART
		, output USART_TX_INT
		, output USART_UDRE_INT
		, output USART_RX_INT
		, input RST_USART_TX_INT
`endif
`ifdef IOR_HAVE_TWI
		, input  RST_TWI_SLAVE
		, output TWI_SLAVE
`endif
`ifdef SBRI_TUNE
		, input [2:0] DO_BIT_SEL
		, output DO_BIT_OUT
`endif
		);
  
//`define	 PORTA_ADDR  6'h00   // PINA 0x00 , DDRA, PORTA
//`define	 PORTB_ADDR  6'h04   // PINB 0x04 , DDRB, PORTB
//`define	 PORTC_ADDR  6'h1b   // PINC 0x1b , DDRC, PORTC
//`define	 RAMWIN_ADDR 6'h1f   // RAMDR 0x1f , RAMAR 0x20
//`define	 TIM0_ADDR   6'h15   // OCR0B 0x15 - TCCR0A 0x19
//`define	 SREG_ADDR   6'h3f   // SREG 0x3f , SPH,SPL,CCP,RSRFR,MCUCR 0x3a
//`define	 MCUCR_ADDR  6'h3a   // MCUCR 0x3a
//`define	 GIFR_ADDR   6'h0b   // GIFR 0x0b , GIMSK 0x0c
//`define	 TIFR_ADDR   6'h25   // TIFR 0x25 , TIMSK 0x26  
//`define	 USART_ADDR  6'h0e   // USART : (0x0e - 0x14 ADC space)
//`define	 SPI_ADDR   6'h2e    // SPI : SPDR 0x2e, SPSR, SPCR
//`define	 TWI_ADDR   6'h28    // TWI : TWSD 0x28, TWSAM,TWSA, TWSSRA

//`define	 INT0_BIT  18     // PC2
//`define	 OC0A_BIT  16     // PC0
//`define	 OC0B_BIT   5     // PA5
//`define	 SS_BIT    16     // PC0
//`define	 SCK_BIT   17     // PC1
//`define	 MOSI_BIT  20     // PC4
//`define	 MISO_BIT  18     // PC2
//`define	 RX_BIT    21     // PC5
//`define	 TX_BIT    22     // PC6
//`define	 XCK_BIT   23     // PC7
//`define	 SDA_BIT   20     // PC4
//`define	 SCL_BIT   17     // PC1

wire  [7:0] DO_async;

`ifdef OUTPUT_MUX_TRI
      assign DO[7:0] = DO_async[7:0];
`else
  `ifdef IOR_DO_SYNC
      reg  [7:0] r_do;
      assign DO[7:0] = r_do[7:0];

    `ifdef SBRI_TUNE
	reg  r_sbri_bit;
	assign DO_BIT_OUT = r_sbri_bit;
	wire v_sbri_bit = (DO_BIT_SEL == 0) ? DO_async[0]
			: (DO_BIT_SEL == 1) ? DO_async[1] 
			: (DO_BIT_SEL == 2) ? DO_async[2]
			: (DO_BIT_SEL == 3) ? DO_async[3]
			: (DO_BIT_SEL == 4) ? DO_async[4]
			: (DO_BIT_SEL == 5) ? DO_async[5]
			: (DO_BIT_SEL == 6) ? DO_async[6]
			: DO_async[7];
    `endif
    `ifdef IOR_DRIVEN_CLK2X
      always @(posedge CLK2X)
      begin
	   if (~CLK) r_do <= DO_async;
       `ifdef SBRI_TUNE
	   if (~CLK) r_sbri_bit <= v_sbri_bit;
       `endif
      end
    `else
      always @(CLK or DO_async)
      begin
	   if (~CLK) r_do = DO_async;
       `ifdef SBRI_TUNE
	   if (~CLK) r_sbri_bit <= v_sbri_bit;
       `endif
      end
    `endif
  `else
      assign DO[7:0] = DO_async[7:0];
  `endif
`endif

`ifdef IOR_HAVE_SREG_SP
    reg  [7:0] r_sreg;
    reg  [7:0] r_sph;
    reg  [7:0] r_spl;

    assign FLAGS_OUT = r_sreg;
    assign SP_OUT  = { r_sph , r_spl };

  `ifdef IOR_LARGE_MAPPER
  `else
    wire OE_sreg = RD & (ADDR == `SREG_ADDR);
    wire OE_sph = RD & (ADDR == (`SREG_ADDR -1));
    wire OE_spl = RD & (ADDR == (`SREG_ADDR -2));  
  `endif 
    wire WE_sreg = WR & (ADDR == `SREG_ADDR);
    wire WE_sph = WR & (ADDR == (`SREG_ADDR -1));
    wire WE_spl = WR & (ADDR == (`SREG_ADDR -2));
`else // IOR_HAVE_SREG_SP 
    wire  [7:0] r_sreg = 0;
    wire  [7:0] r_sph = 0;
    wire  [7:0] r_spl = 0;

    wire OE_sreg = 0;
    wire OE_sph = 0;
    wire OE_spl = 0;
`endif

`ifdef IOR_HAVE_RAMWIN_HA
    reg  [2:0] r_ramar_h;
`else
    wire  [2:0] r_ramar_h = 0;
`endif
`ifdef IOR_HAVE_RAMWIN
    reg  [7:0] r_ramar;

  `ifdef IOR_LARGE_MAPPER
  `else
    wire OE_ramar = RD & (ADDR == `RAMWIN_ADDR); 
    wire OE_ramdr = RD & (ADDR == (`RAMWIN_ADDR +1)); 
  `endif
    wire WE_ramar = WR & (ADDR == `RAMWIN_ADDR);
    wire WE_ramdr = WR & (ADDR == (`RAMWIN_ADDR +1));
    
    assign RAM_ADDR[7:0] = r_ramar;
    assign RAM_ADDR[10:8] = r_ramar_h;
    assign RAM_WE = WE_ramdr;
    assign RAM_DI = DO;
`endif

`ifdef IOR_HAVE_TIM0
  `ifdef IOR_LARGE_MAPPER
  `else
    wire OE_tifr = RD & (ADDR == `TIFR_ADDR); 
    wire OE_timsk = RD & (ADDR == (`TIFR_ADDR +1)); 
  `endif
    wire WE_tifr = WR & (ADDR == `TIFR_ADDR);
    wire WE_timsk = WR & (ADDR == (`TIFR_ADDR +1));

    reg t0_ovf_prev;  
    reg t0_compa_prev;
    reg t0_compb_prev;  
    reg t0_ovf, t0_ovf_ie; // bit0
    reg t0_compa, t0_compa_ie; // bit1
    reg t0_compb, t0_compb_ie; // bit2

    assign TIM0_OVF   = t0_ovf & t0_ovf_ie;       // int 10
    assign TIM0_COMPA = t0_compa & t0_compa_ie ;  // int 11
    assign TIM0_COMPB = t0_compb & t0_compb_ie ;  // int 12
  `ifdef TIM0_EXPORT_PS
     wire [11:0] TIM0_PS;
  `endif
`else // IOR_HAVE_TIM0
    wire OE_tifr = 0; 
    wire WE_tifr = 0;
    wire OE_timsk = 0; 
    wire WE_timsk = 0;
`endif


 //reg intf0, int0; // bit0
`ifdef IOR_HAVE_INT0
  `ifdef IOR_LARGE_MAPPER
  `else
    wire OE_gifr = RD & (ADDR == `GIFR_ADDR); 
    wire OE_gimsk = RD & (ADDR == (`GIFR_ADDR +1));
    wire OE_mcucr = RD & (ADDR == `MCUCR_ADDR); 
  `endif
    wire WE_gifr = WR & (ADDR == `GIFR_ADDR);
    wire WE_gimsk = WR & (ADDR == (`GIFR_ADDR +1));
    wire WE_mcucr = WR & (ADDR == `MCUCR_ADDR);

    reg gi_int0_prev; 
    reg gi_int0, gi_int0_ie; // bit0
    reg [1:0] mcucr_isc;
  
    assign INT0 = gi_int0 & gi_int0_ie;
  `ifdef IOR_HAVE_INT1
      reg gi_int1_prev; 
      reg gi_int1, gi_int1_ie; // bit1
      reg [1:0] mcucr_isc1;
  
      assign INT1 = gi_int1 & gi_int1_ie;
  `else
      wire gi_int1 = 0;
      wire gi_int1_ie = 0;
  `endif
`else // IOR_HAVE_INT0
    wire OE_gifr = 0; 
    wire WE_gifr = 0;
    wire OE_gimsk = 0; 
    wire WE_gimsk = 0;
    wire OE_mcucr = 0; 
    wire WE_mcucr = 0;
`endif

    wire [23:0] PORT_IN;
`ifdef IOR_PORT_ESCALATION
`else
    wire [23:0] PIN;
    wire [23:0] DDR;
`endif

`ifdef IOR_HAVE_SPI
    wire SPI_MASTER;
    wire SPI_SLAVE;
    wire SS_DDR;
    wire SS_IN;
    wire SCK_IN;
    wire MOSI_IN;
    wire MISO_IN;
`endif
`ifdef IOR_HAVE_USART
    wire XCK_DDR;
    wire XCK_IN;
    wire XCK_OUT;
    wire TX_OUT;
    wire RX_IN;
    wire USART_XCKEN;
    wire USART_TXEN;
`endif
`ifdef IOR_HAVE_TWI
    wire TWI_EN;
    wire SDA_OUT;
    wire SCL_OUT;
    wire SDA_IN;
    wire SCL_IN;
`endif

`ifdef IOR_LARGE_MAPPER
`else
`ifdef OUTPUT_MUX_TRI
`else
    wire [7:0] pa_DO;
    wire [7:0] pb_DO;
    wire [7:0] pc_DO;
    wire [7:0] t0_DO;
    wire [7:0] spi_DO;
    wire [7:0] usart_DO;
`endif
`endif

`ifdef IOR_LARGE_MAPPER
  `ifdef IOR_HAVE_USART
	wire WE_ubrrl = WR & (ADDR == `USART_ADDR);
	wire WE_ubrrh = WR & (ADDR == (`USART_ADDR + 1));
	wire WE_ucsrc = WR & (ADDR == (`USART_ADDR + 2));
	wire WE_ucsrb = WR & (ADDR == (`USART_ADDR + 3));
	wire WE_ucsra = WR & (ADDR == (`USART_ADDR + 4));
	wire WE_udr   = WR & (ADDR == (`USART_ADDR + 5));
	wire OE_udr   = RD & (ADDR == (`USART_ADDR + 5)); 
	wire [7:0] DO_ubrrl;
	wire [7:0] DO_ubrrh;
	wire [7:0] DO_ucsrc;
	wire [7:0] DO_ucsrb;
	wire [7:0] DO_ucsra;
	wire [7:0] DO_udr;
  `endif
  `ifdef IOR_HAVE_SPI
	wire WE_spdr = WR & (ADDR == `SPI_ADDR);
	wire WE_spsr = WR & (ADDR == (`SPI_ADDR + 1));
	wire WE_spcr = WR & (ADDR == (`SPI_ADDR + 2));
	wire OE_spdr = RD & (ADDR == `SPI_ADDR);
	wire OE_spsr = RD & (ADDR == (`SPI_ADDR + 1)); 
	wire [7:0] DO_spdr;
	wire [7:0] DO_spsr;
	wire [7:0] DO_spcr;
  `endif
  `ifdef IOR_HAVE_PORTA
	wire WE_ddr_a   = WR & (ADDR == `PORTA_ADDR);
	wire WE_pin_a   = WR & (ADDR == `PORTA_ADDR +1);
	wire WE_port_a  = WR & (ADDR == `PORTA_ADDR +2);
	wire [7:0] DO_ddr_a;
	wire [7:0] DO_pin_a;
	wire [7:0] DO_port_a;
  `endif
  `ifdef IOR_HAVE_PORTB
	wire WE_ddr_b   = WR & (ADDR == `PORTB_ADDR);
	wire WE_pin_b   = WR & (ADDR == `PORTB_ADDR +1);
	wire WE_port_b  = WR & (ADDR == `PORTB_ADDR +2);
	wire [7:0] DO_ddr_b;
	wire [7:0] DO_pin_b;
	wire [7:0] DO_port_b;
  `endif
  `ifdef IOR_HAVE_PORTC
	wire WE_ddr_c   = WR & (ADDR == `PORTC_ADDR);
	wire WE_pin_c   = WR & (ADDR == `PORTC_ADDR +1);
	wire WE_port_c  = WR & (ADDR == `PORTC_ADDR +2);
	wire [7:0] DO_ddr_c;
	wire [7:0] DO_pin_c;
	wire [7:0] DO_port_c;
  `endif
  `ifdef IOR_HAVE_TIM0
	wire WE_ocr0b = WR & (ADDR == `TIM0_ADDR);
	wire WE_ocr0a = WR & (ADDR == (`TIM0_ADDR + 1));
	wire WE_tcnt0 = WR & (ADDR == (`TIM0_ADDR + 2));
	wire WE_tccr0b = WR & (ADDR == (`TIM0_ADDR + 3));
	wire WE_tccr0a = WR & (ADDR == (`TIM0_ADDR + 4));
	wire [7:0] DO_ocr0b;
	wire [7:0] DO_ocr0a;
	wire [7:0] DO_tcnt0;
	wire [7:0] DO_tccr0b;
	wire [7:0] DO_tccr0a;
	wire [7:0] DO_tifr = {5'b0 , t0_compb, t0_compa ,t0_ovf};
	wire [7:0] DO_timsk = {5'b0 , t0_compb_ie, t0_compa_ie ,t0_ovf_ie};
  `endif
  `ifdef IOR_HAVE_INT0
	wire [7:0] DO_gifr  = {7'b0 , gi_int0};
	wire [7:0] DO_gimsk = {1'b0 , r_ramar_h , 2'b0, gi_int1_ie, gi_int0_ie};
	wire [7:0] DO_mcucr = {mcucr_isc , 6'b0 };
  `endif

  assign DO_async = ~RD ? 0
	: (ADDR == `SREG_ADDR) ?      r_sreg 
	: (ADDR == (`SREG_ADDR -1)) ? r_sph
	: (ADDR == (`SREG_ADDR -2)) ? r_spl
  `ifdef IOR_HAVE_USART
	: (ADDR == `USART_ADDR) ?       DO_ubrrl
	: (ADDR == (`USART_ADDR + 1)) ? DO_ubrrh
	: (ADDR == (`USART_ADDR + 2)) ? DO_ucsrc
	: (ADDR == (`USART_ADDR + 3)) ? DO_ucsrb
	: (ADDR == (`USART_ADDR + 4)) ? DO_ucsra
	: (ADDR == (`USART_ADDR + 5)) ? DO_udr
  `endif
  `ifdef IOR_HAVE_SPI
	: (ADDR == `SPI_ADDR) ?       DO_spdr
	: (ADDR == (`SPI_ADDR + 1)) ? DO_spsr
	: (ADDR == (`SPI_ADDR + 2)) ? DO_spcr
  `endif
  `ifdef IOR_HAVE_PORTA
	: (ADDR == `PORTA_ADDR) ?     DO_ddr_a
	: (ADDR == `PORTA_ADDR +1) ?  DO_pin_a
	: (ADDR == `PORTA_ADDR +2) ?  DO_port_a
  `endif
  `ifdef IOR_HAVE_PORTB
	: (ADDR == `PORTB_ADDR) ?     DO_ddr_b
	: (ADDR == `PORTB_ADDR +1) ?  DO_pin_b
	: (ADDR == `PORTB_ADDR +2) ?  DO_port_b
  `endif
  `ifdef IOR_HAVE_PORTC
	: (ADDR == `PORTC_ADDR) ?     DO_ddr_c
	: (ADDR == `PORTC_ADDR +1) ?  DO_pin_c
	: (ADDR == `PORTC_ADDR +2) ?  DO_port_c
  `endif
  `ifdef IOR_HAVE_TIM0
	: (ADDR == `TIM0_ADDR) ?       DO_ocr0b
	: (ADDR == (`TIM0_ADDR + 1)) ? DO_ocr0a
	: (ADDR == (`TIM0_ADDR + 2)) ? DO_tcnt0
	: (ADDR == (`TIM0_ADDR + 3)) ? DO_tccr0b
	: (ADDR == (`TIM0_ADDR + 4)) ? DO_tccr0a
	: (ADDR == `TIFR_ADDR) ?       DO_tifr
	: (ADDR == (`TIFR_ADDR +1)) ?  DO_timsk
  `endif
  `ifdef IOR_HAVE_RAMWIN
	: (ADDR == `RAMWIN_ADDR)      ? r_ramar
	: (ADDR == (`RAMWIN_ADDR +1)) ? RAM_DO
  `endif
  `ifdef IOR_HAVE_INT0
	: (ADDR == `GIFR_ADDR) ?        DO_gifr
	: (ADDR == (`GIFR_ADDR +1)) ?   DO_gimsk
	: (ADDR == `MCUCR_ADDR) ?       DO_mcucr
  `endif
	: 0;

`endif // IOR_LARGE_MAPPER

`ifdef IOR_HAVE_SPI
    rtavr_ior_spi #(.BASE_ADDR(`SPI_ADDR)) i_spi(
		.CLK(CLK), .RESET(RESET), .DI(DI)
`ifdef IOR_LARGE_MAPPER
		, .WE_spdr(WE_spdr)
		, .WE_spsr(WE_spsr)
		, .WE_spcr(WE_spcr)
		, .OE_spdr(OE_spdr)
		, .OE_spsr(OE_spsr)
		, .DO_spdr(DO_spdr)
		, .DO_spsr(DO_spsr)
		, .DO_spcr(DO_spcr)
`else
  `ifdef OUTPUT_MUX_TRI
		, .DO(DO)
  `else
		, .DO(spi_DO)
  `endif
		, .WR(WR), .RD(RD), .ADDR(ADDR) 
`endif
		, .SS_DDR(SS_DDR), .SS_IN(SS_IN)
		, .SCK_OUT(SCK_OUT), .SCK_IN(SCK_IN)
		, .MOSI_OUT(MOSI_OUT), .MOSI_IN(MOSI_IN)
		, .MISO_OUT(MISO_OUT), .MISO_IN(MISO_IN)
		, .SPI_MASTER(SPI_MASTER), .SPI_SLAVE(SPI_SLAVE)
		, .SPI_INT(SPI_INT)
		, .RST_SPI_INT(RST_SPI_INT)
		);
`endif
`ifdef IOR_HAVE_USART
    rtavr_ior_usart #(.BASE_ADDR(`USART_ADDR)) i_usart(
		.CLK(CLK), .RESET(RESET), .DI(DI)
`ifdef IOR_LARGE_MAPPER
		, .WE_ubrrl(WE_ubrrl)
		, .WE_ubrrh(WE_ubrrh)
		, .WE_ucsrc(WE_ucsrc)
		, .WE_ucsrb(WE_ucsrb)
		, .WE_ucsra(WE_ucsra)
		, .WE_udr(WE_udr)
		, .OE_udr(OE_udr)
		, .DO_ubrrl(DO_ubrrl)
		, .DO_ubrrh(DO_ubrrh)
		, .DO_ucsrc(DO_ucsrc)
		, .DO_ucsrb(DO_ucsrb)
		, .DO_ucsra(DO_ucsra)
		, .DO_udr(DO_udr)
`else
  `ifdef OUTPUT_MUX_TRI
		, .DO(DO)
  `else
		, .DO(usart_DO)
  `endif
		, .WR(WR), .RD(RD), .ADDR(ADDR) 
`endif

		, .XCK_DDR(XCK_DDR), .XCK_OUT(XCK_OUT), .XCK_IN(XCK_IN) 
		, .TX_OUT(TX_OUT), .RX_IN(RX_IN)
		, .TXEN(USART_TXEN), .XCKEN(USART_XCKEN)

		, .TX_INT(USART_TX_INT)
		, .UDRE_INT(USART_UDRE_INT)
		, .RX_INT(USART_RX_INT)
		, .RST_TX_INT(RST_USART_TX_INT)
		);
`endif
`ifdef IOR_HAVE_TWI
    rtavr_ior_twi #(.BASE_ADDR(`TWI_ADDR)) i_twi(
		.CLK(CLK), .RESET(RESET), .DI(DI)
  `ifdef OUTPUT_MUX_TRI
		, .DO(DO)
  `else
		, .DO(twi_DO)
  `endif
		, .WR(WR), .RD(RD), .ADDR(ADDR) 
		, .TWI_EN(TWI_EN)
		, .SDA_OUT(SDA_OUT), .SDA_IN(SDA_IN)
		, .SCL_OUT(SCL_OUT), .SCL_IN(SCL_IN)
		, .TWI_SLAVE(TWI_SLAVE)
		, .RST_TWI_SLAVE(RST_TWI_SLAVE)
		);
`endif
`ifdef IOR_HAVE_PORTA
    rtavr_ior_port #(.BASE_ADDR(`PORTA_ADDR)) i_port_a(
		.CLK(CLK), .RESET(RESET), .DI(DI)
`ifdef IOR_LARGE_MAPPER
		, .WE_ddr(WE_ddr_a)
		, .WE_pin(WE_pin_a)
		, .WE_port(WE_port_a)
		, .DO_ddr(DO_ddr_a)
		, .DO_pin(DO_pin_a)
		, .DO_port(DO_port_a)
`else
  `ifdef OUTPUT_MUX_TRI
		, .DO(DO)
  `else
		, .DO(pa_DO)
  `endif
		, .WR(WR), .RD(RD), .ADDR(ADDR)
`endif
		, .PIN(PIN[7:0]), .PORT(PORT_IN[7:0]), .DDR(DDR[7:0])
		);
`endif
`ifdef IOR_HAVE_PORTB
    rtavr_ior_port #(.BASE_ADDR(`PORTB_ADDR)) i_port_b(
		.CLK(CLK), .RESET(RESET), .DI(DI)
`ifdef IOR_LARGE_MAPPER
		, .WE_ddr(WE_ddr_b)
		, .WE_pin(WE_pin_b)
		, .WE_port(WE_port_b)
		, .DO_ddr(DO_ddr_b)
		, .DO_pin(DO_pin_b)
		, .DO_port(DO_port_b)
`else
  `ifdef OUTPUT_MUX_TRI
		, .DO(DO)
  `else
		, .DO(pb_DO)
  `endif
		, .WR(WR), .RD(RD), .ADDR(ADDR)
`endif
		, .PIN(PIN[15:8]), .PORT(PORT_IN[15:8]), .DDR(DDR[15:8])
		);
`endif
`ifdef IOR_HAVE_PORTC
    rtavr_ior_port #(.BASE_ADDR(`PORTC_ADDR)) i_port_c(
		.CLK(CLK), .RESET(RESET), .DI(DI)
`ifdef IOR_LARGE_MAPPER
		, .WE_ddr(WE_ddr_c)
		, .WE_pin(WE_pin_c)
		, .WE_port(WE_port_c)
		, .DO_ddr(DO_ddr_c)
		, .DO_pin(DO_pin_c)
		, .DO_port(DO_port_c)
`else
  `ifdef OUTPUT_MUX_TRI
		, .DO(DO)
  `else
		, .DO(pc_DO)
  `endif
		, .WR(WR), .RD(RD), .ADDR(ADDR)
`endif
		, .PIN(PIN[23:16]), .PORT(PORT_IN[23:16]), .DDR(DDR[23:16])
		);
`endif

`ifdef IOR_HAVE_TIM0
    rtavr_ior_timer0 #(.BASE_ADDR(`TIM0_ADDR)) i_timer0(
		.CLK(CLK), .RESET(RESET), .DI(DI)
`ifdef IOR_LARGE_MAPPER
		, .WE_ocr0b(WE_ocr0b)
		, .WE_ocr0a(WE_ocr0a)
		, .WE_tcnt0(WE_tcnt0)
		, .WE_tccr0b(WE_tccr0b)
		, .WE_tccr0a(WE_tccr0a)
		, .DO_ocr0b(DO_ocr0b)
		, .DO_ocr0a(DO_ocr0a)
		, .DO_tcnt0(DO_tcnt0)
		, .DO_tccr0b(DO_tccr0b)
		, .DO_tccr0a(DO_tccr0a)
`else
  `ifdef OUTPUT_MUX_TRI
		, .DO(DO)
  `else
		, .DO(t0_DO)
  `endif
		, .WR(WR), .RD(RD), .ADDR(ADDR)
`endif
`ifdef IOR_HAVE_TIM0_OC0A
		, .OC0A_EN(OC0A_EN)
		, .OC0A_OUT(OC0A_OUT)
`endif
`ifdef IOR_HAVE_TIM0_OC0B
		, .OC0B_EN(OC0B_EN)
		, .OC0B_OUT(OC0B_OUT)
`endif
		, .TIM0_OVF(TIM0_OVF_IN)
		, .TIM0_COMPB(TIM0_COMPB_IN)
		, .TIM0_COMPA(TIM0_COMPA_IN)
    `ifdef TIM0_EXPORT_PS
		, .PS(TIM0_PS)
    `endif
		 );
`endif

`ifdef IOR_LARGE_MAPPER
`else // IOR_LARGE_MAPPER

`ifdef OUTPUT_MUX_TRI
`elsif OUTPUT_MUX_OR
`else
  
  `ifdef IOR_HAVE_SPI
    wire OE_spi = RD & ( (ADDR == `SPI_ADDR) | (ADDR == `SPI_ADDR+1)
		     | (ADDR == `SPI_ADDR+2) );
  `else
    assign spi_DO = 8'h00;
    wire OE_spi = 0;
  `endif

  `ifdef IOR_HAVE_USART
    wire OE_usart = RD & ( (ADDR == `USART_ADDR) | (ADDR == `USART_ADDR+1)
		     | (ADDR == `USART_ADDR+2) | (ADDR == `USART_ADDR+3)
		     | (ADDR == `USART_ADDR+4) | (ADDR == `USART_ADDR+5) );
  `else
    assign usart_DO = 8'h00;
    wire OE_usart = 0;
  `endif

  `ifdef IOR_HAVE_TWI
    wire OE_twi = RD & ( (ADDR == `TWI_ADDR) | (ADDR == `TWI_ADDR+1)
		     | (ADDR == `TWI_ADDR+2) | (ADDR == `TWI_ADDR+3) 
		     | (ADDR == `TWI_ADDR+4) | (ADDR == `TWI_ADDR+5) );
  `else
    assign twi_DO = 8'h00;
    wire OE_twi = 0;
  `endif

  `ifdef IOR_HAVE_PORTA
    wire OE_pa = RD & ( (ADDR == `PORTA_ADDR) | (ADDR == `PORTA_ADDR+1)
		     | (ADDR == `PORTA_ADDR+2) );
  `else
    assign pa_DO = 8'h00;
    wire OE_pa = 0;
  `endif
  `ifdef IOR_HAVE_PORTB
    wire OE_pb = RD & ( (ADDR == `PORTB_ADDR) | (ADDR == `PORTB_ADDR+1)
		     | (ADDR == `PORTB_ADDR+2) );
  `else
    assign pb_DO = 8'h00;
    wire OE_pb = 0;
  `endif

  `ifdef IOR_HAVE_PORTC
    wire OE_pc = RD & ( (ADDR == `PORTC_ADDR) | (ADDR == `PORTC_ADDR+1)
		     | (ADDR == `PORTC_ADDR+2) );
  `else
    assign pc_DO = 8'h00;
    wire OE_pc = 0;
  `endif

  `ifdef IOR_HAVE_TIM0
   wire OE_t0 = RD & ( (ADDR == `TIM0_ADDR) | (ADDR == `TIM0_ADDR+1)
                     | (ADDR == `TIM0_ADDR+2) | (ADDR == `TIM0_ADDR+3)
		     | (ADDR == `TIM0_ADDR+4) );

  `else
    assign t0_DO = 8'h00;
    wire OE_t0 = 0;
  `endif
`endif //OUTPUT_MUX_TRI

   wire [7:0] i_DO = OE_sreg ? r_sreg : OE_sph ? r_sph : OE_spl ? r_spl
  `ifdef IOR_HAVE_TIM0
                  : OE_tifr ? {5'b0 , t0_compb, t0_compa ,t0_ovf}
		  : OE_timsk ? {5'b0 , t0_compb_ie, t0_compa_ie ,t0_ovf_ie}
  `endif
  `ifdef IOR_HAVE_RAMWIN
		  : OE_ramar ? r_ramar
		  : OE_ramdr ? RAM_DO
  `endif
  `ifdef IOR_HAVE_INT0
                  : OE_gifr ? {7'b0 , gi_int0}
		  : OE_gimsk ? {1'b0 , r_ramar_h , 2'b0, gi_int1_ie, gi_int0_ie}
		  : OE_mcucr ? {mcucr_isc , 6'b0 }
  `endif

		  : 0;

`ifdef OUTPUT_MUX_TRI
   bufif1(DO_async, i_DO, (OE_sreg | OE_sph | OE_spl | OE_tifr | OE_timsk ));
`elsif OUTPUT_MUX_OR
   assign DO_async =  (OE_sreg | OE_sph | OE_spl | OE_tifr | OE_timsk ) ? i_DO
              : pa_DO | pb_DO | pc_DO | t0_DO | spi_DO | usart_DO;   
`else
   assign DO_async =  (OE_sreg | OE_sph | OE_spl | OE_tifr | OE_timsk ) ? i_DO
              : OE_pa ? pa_DO : OE_pb ? pb_DO : OE_pc ? pc_DO
	      : OE_t0 ? t0_DO : OE_spi ? spi_DO 
	      : OE_usart ? usart_DO : 8'h00; 
`endif
`endif // IOR_LARGE_MAPPER


`ifdef IOR_HAVE_SREG_SP
  `ifdef IOR_DRIVEN_CLK2X
    always @(posedge CLK)
  `else
    always @(posedge CLK2X)
  `endif
    begin
      if (RESET == 0)
        begin
	r_sreg <= 0;
  `ifdef FIXED_SPH
	r_sph <= `FIXED_SPH;
  `else
	r_sph <= 0;
  `endif
	r_spl <= 0;
        end
  `ifdef IOR_DRIVEN_CLK2X
      else if ( CLK )
  `else
      else
  `endif
        begin
	if(WE_sreg) r_sreg <= DI;
	else r_sreg <= FLAGS_IN;

  `ifdef FIXED_SPH
  `else
    `ifdef IOR_SPHBITS
	if(WE_sph) r_sph <= DI[`IOR_SPHBITS-1:0];
	else r_sph <= SP_IN[`IOR_SPHBITS+7:8];
    `else
	if(WE_sph) r_sph <= DI;
	else r_sph <= SP_IN[15:8];
    `endif
  `endif
	if(WE_spl) r_spl <= DI;
	else r_spl <= SP_IN[7:0];
        end
    end
`endif

  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
`ifdef IOR_HAVE_TIM0
	t0_compb_ie <= 0;
	t0_compa_ie <= 0;
	t0_ovf_ie <= 0;

	t0_ovf <= 0;
	t0_compa <= 0;

	t0_compb <= 0;

	t0_ovf_prev   <=  0;
	t0_compa_prev <=  0;
	t0_compb_prev <=  0;
`endif

`ifdef IOR_HAVE_RAMWIN
	r_ramar	   <= 0;
  `ifdef IOR_HAVE_RAMWIN_HA
	r_ramar_h  <= 0;
  `endif
`endif
`ifdef IOR_HAVE_INT0
	mcucr_isc <= 0;

	gi_int0_ie <= 0;
	gi_int0 <= 0;
	gi_int0_prev   <= 0;
  `ifdef IOR_HAVE_INT1
	mcucr_isc1 <= 0;

	gi_int1_ie <= 0;
	gi_int1 <= 0;
	gi_int1_prev   <= 0;
  `endif
`endif
      end
    else
      begin
`ifdef IOR_HAVE_TIM0
	if(WE_timsk) {t0_compb_ie, t0_compa_ie ,t0_ovf_ie} <= DI[2:0];

	if(RST_TIM0_OVF) t0_ovf <= 0;
	else if((WE_tifr & DI[0]) 
             | (t0_ovf_prev ^ TIM0_OVF_IN) & ~t0_ovf_prev ) t0_ovf <= 1;

	if(RST_TIM0_COMPA) t0_compa <= 0;
	else if((WE_tifr & DI[1])
             | (t0_compa_prev ^ TIM0_COMPA_IN) & ~t0_compa_prev) t0_compa <= 1;

	if(RST_TIM0_COMPB) t0_compb <= 0;
	else if((WE_tifr & DI[2])
             | (t0_compb_prev ^ TIM0_COMPB_IN) & ~t0_compb_prev) t0_compb <= 1;

	t0_ovf_prev   <=  TIM0_OVF_IN;
	t0_compa_prev <=  TIM0_COMPA_IN;
	t0_compb_prev <=  TIM0_COMPB_IN; 
`endif //  IOR_HAVE_TIM0

`ifdef IOR_HAVE_RAMWIN
	if(WE_ramar) r_ramar	   <= DI[7:0];
  `ifdef IOR_HAVE_RAMWIN_HA
	if(WE_gimsk) r_ramar_h	   <= DI[6:4];
  `endif
`endif
`ifdef IOR_HAVE_INT0
	if(WE_mcucr) {mcucr_isc} <= DI[7:6];

	if(WE_gimsk) {gi_int0_ie} <= DI[0];
	if(RST_INT0 | (WE_gifr & DI[0])) gi_int0 <= 0;
	else if((gi_int0_prev ^ PIN[`INT0_BIT])
                 & ((mcucr_isc == 1) 
                   | (mcucr_isc == 2) & ( gi_int0_prev) 
                   | (mcucr_isc == 3) & ~( gi_int0_prev) )  
	       | ((mcucr_isc == 0) & ~PIN[`INT0_BIT]) ) gi_int0 <= 1;

	gi_int0_prev   <= PIN[`INT0_BIT];
  `ifdef IOR_HAVE_INT1
	if(WE_mcucr) {mcucr_isc1} <= DI[5:4];

	if(WE_gimsk) {gi_int1_ie} <= DI[1];
	if(RST_INT1 | (WE_gifr & DI[1])) gi_int1 <= 0;
	else if((gi_int1_prev ^ PIN[`INT1_BIT])
                 & ((mcucr_isc1 == 1) 
                   | (mcucr_isc1 == 2) & ( gi_int1_prev) 
                   | (mcucr_isc1 == 3) & ~( gi_int1_prev) )  
	       | ((mcucr_isc1 == 0) & ~PIN[`INT1_BIT]) ) gi_int1 <= 1;

	gi_int1_prev   <= PIN[`INT1_BIT];
  `endif
`endif // IOR_HAVE_INT0
      end
  end

`ifdef IOR_PORT_ESCALATION

  generate 
    genvar i;
  `ifdef IOR_HAVE_PORTA
    for (i=0; i<8; i=i+1)
       begin : port_a_out
	assign PORT[i] = 
    `ifdef IOR_HAVE_TIM0_OC0A
		(`OC0A_BIT == i) & OC0A_EN ? OC0A_OUT :
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
		(`OC0B_BIT == i) & OC0B_EN ? OC0B_OUT :
    `endif
    `ifdef IOR_HAVE_SPI
		(`SCK_BIT == i) & SPI_MASTER ? SCK_OUT :
		(`MOSI_BIT == i) & SPI_MASTER ? MOSI_OUT :
		(`MISO_BIT == i) & SPI_SLAVE ? MISO_OUT :
    `endif
    `ifdef IOR_HAVE_USART
		(`XCK_BIT == i) & USART_XCKEN ? XCK_OUT :
		(`TX_BIT == i) & USART_TXEN ? TX_OUT :
    `endif
    `ifdef IOR_HAVE_TWI
		(`SDA_BIT == i) & TWI_EN & ~SDA_OUT ? 0 :
		(`SCL_BIT == i) & TWI_EN & ~SCL_OUT ? 0 :
    `endif
		PORT_IN[i];
       end
  `endif
  `ifdef IOR_HAVE_PORTB
    for (i=8; i<16; i=i+1)
       begin : port_b_out
	assign PORT[i] = 
    `ifdef IOR_HAVE_TIM0_OC0A
		(`OC0A_BIT == i) & OC0A_EN ? OC0A_OUT :
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
		(`OC0B_BIT == i) & OC0B_EN ? OC0B_OUT :
    `endif
    `ifdef IOR_HAVE_SPI
		(`SCK_BIT == i) & SPI_MASTER ? SCK_OUT :
		(`MOSI_BIT == i) & SPI_MASTER ? MOSI_OUT :
		(`MISO_BIT == i) & SPI_SLAVE ? MISO_OUT :
    `endif
    `ifdef IOR_HAVE_USART
		(`XCK_BIT == i) & USART_XCKEN ? XCK_OUT :
		(`TX_BIT == i) & USART_TXEN ? TX_OUT :
    `endif
    `ifdef IOR_HAVE_TWI
		(`SDA_BIT == i) & TWI_EN & ~SDA_OUT ? 0 :
		(`SCL_BIT == i) & TWI_EN & ~SCL_OUT ? 0 :
    `endif
		PORT_IN[i];
       end
  `endif
  `ifdef IOR_HAVE_PORTC
    for (i=16; i<24; i=i+1)
       begin : port_c_out
	assign PORT[i] = 
    `ifdef IOR_HAVE_TIM0_OC0A
		(`OC0A_BIT == i) & OC0A_EN ? OC0A_OUT :
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
		(`OC0B_BIT == i) & OC0B_EN ? OC0B_OUT :
    `endif
    `ifdef IOR_HAVE_SPI
		(`SCK_BIT == i) & SPI_MASTER ? SCK_OUT :
		(`MOSI_BIT == i) & SPI_MASTER ? MOSI_OUT :
		(`MISO_BIT == i) & SPI_SLAVE ? MISO_OUT :
    `endif
    `ifdef IOR_HAVE_USART
		(`XCK_BIT == i) & USART_XCKEN ? XCK_OUT :
		(`TX_BIT == i) & USART_TXEN ? TX_OUT :
    `endif
    `ifdef IOR_HAVE_TWI
		(`SDA_BIT == i) & TWI_EN & ~SDA_OUT ? 0 :
		(`SCL_BIT == i) & TWI_EN & ~SCL_OUT ? 0 :
    `endif
		PORT_IN[i];
       end
  `endif

  endgenerate
`else // IOR_PORT_ESCALATION

  generate 
    genvar i;
  `ifdef IOR_HAVE_PORTA
    for (i=0; i<8; i=i+1)
       begin : port_a_out
	assign PORTA[i] = DDR[i] ? 
    `ifdef IOR_HAVE_TIM0_OC0A
		(`OC0A_BIT == i) & OC0A_EN ? OC0A_OUT :
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
		(`OC0B_BIT == i) & OC0B_EN ? OC0B_OUT :
    `endif
    `ifdef IOR_HAVE_SPI
		(`SCK_BIT == i) & SPI_MASTER ? SCK_OUT :
		(`MOSI_BIT == i) & SPI_MASTER ? MOSI_OUT :
		(`MISO_BIT == i) & SPI_SLAVE ? MISO_OUT :
    `endif
    `ifdef IOR_HAVE_USART
		(`XCK_BIT == i) & USART_XCKEN ? XCK_OUT :
		(`TX_BIT == i) & USART_TXEN ? TX_OUT :
    `endif
    `ifdef IOR_HAVE_TWI
		(`SDA_BIT == i) & TWI_EN & ~SDA_OUT ? 0 :
		(`SCL_BIT == i) & TWI_EN & ~SCL_OUT ? 0 :
    `endif
		PORT_IN[i]
		: (~DDR[i] & PORT_IN[i]) ? "H"
		: "Z";
	assign PIN[i] = PORTA[i];
       end
  `endif
  `ifdef IOR_HAVE_PORTB
    for (i=8; i<16; i=i+1)
       begin : port_b_out
	assign PORTC[i-16] = DDR[i] ? 
    `ifdef IOR_HAVE_TIM0_OC0A
		(`OC0A_BIT == i) & OC0A_EN ? OC0A_OUT :
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
		(`OC0B_BIT == i) & OC0B_EN ? OC0B_OUT :
    `endif
    `ifdef IOR_HAVE_SPI
		(`SCK_BIT == i) & SPI_MASTER ? SCK_OUT :
		(`MOSI_BIT == i) & SPI_MASTER ? MOSI_OUT :
		(`MISO_BIT == i) & SPI_SLAVE ? MISO_OUT :
    `endif
    `ifdef IOR_HAVE_USART
		(`XCK_BIT == i) & USART_XCKEN ? XCK_OUT :
		(`TX_BIT == i) & USART_TXEN ? TX_OUT :
    `endif
    `ifdef IOR_HAVE_TWI
		(`SDA_BIT == i) & TWI_EN & ~SDA_OUT ? 0 :
		(`SCL_BIT == i) & TWI_EN & ~SCL_OUT ? 0 :
    `endif
		PORT_IN[i]
		: (~DDR[i] & PORT_IN[i]) ? "H"
		: "Z";
	assign PIN[i] = PORTC[i-16];
       end
  `endif
  `ifdef IOR_HAVE_PORTC
    for (i=16; i<24; i=i+1)
       begin : port_c_out
	assign PORTC[i-16] = DDR[i] ? 
    `ifdef IOR_HAVE_TIM0_OC0A
		(`OC0A_BIT == i) & OC0A_EN ? OC0A_OUT :
    `endif
    `ifdef IOR_HAVE_TIM0_OC0B
		(`OC0B_BIT == i) & OC0B_EN ? OC0B_OUT :
    `endif
    `ifdef IOR_HAVE_SPI
		(`SCK_BIT == i) & SPI_MASTER ? SCK_OUT :
		(`MOSI_BIT == i) & SPI_MASTER ? MOSI_OUT :
		(`MISO_BIT == i) & SPI_SLAVE ? MISO_OUT :
    `endif
    `ifdef IOR_HAVE_USART
		(`XCK_BIT == i) & USART_XCKEN ? XCK_OUT :
		(`TX_BIT == i) & USART_TXEN ? TX_OUT :
    `endif
    `ifdef IOR_HAVE_TWI
		(`SDA_BIT == i) & TWI_EN & ~SDA_OUT ? 0 :
		(`SCL_BIT == i) & TWI_EN & ~SCL_OUT ? 0 :
    `endif
		PORT_IN[i]
		: (~DDR[i] & PORT_IN[i]) ? "H"
		: "Z";
	assign PIN[i] = PORTC[i-16];
       end
  `endif
  endgenerate
`endif // IOR_PORT_ESCALATION
  `ifdef IOR_HAVE_SPI
	assign SS_DDR = DDR[`SS_BIT];
	assign SS_IN = PIN[`SS_BIT];
	assign SCK_IN = PIN[`SCK_BIT];
	assign MOSI_IN = PIN[`MOSI_BIT];
	assign MISO_IN = PIN[`MISO_BIT];
  `endif
  `ifdef IOR_HAVE_USART
	assign XCK_DDR = DDR[`XCK_BIT];
	assign XCK_IN = PIN[`XCK_BIT];
	assign RX_IN = PIN[`RX_BIT];
  `endif
  `ifdef IOR_HAVE_TWI
	assign SDA_IN = PIN[`SDA_BIT];
	assign SCL_IN = PIN[`SCL_BIT];
  `endif
endmodule
