/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_TRI 

//`define OUTPUT_MUX_NONE
//`define OUTPUT_MUX_TRI

// Module Name: rtavr_rom
//  ver 005
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                   (1KW)       NORMAL MUX_TRI(FAKE) MUX_NONE  
// Number of Slice Flip Flops:      9          9         9    
// Number of 4 input LUTs:         10         10        10
// Number of occupied Slices:      10         10        10
// Total Number of 4 input LUTs:   10         10        10
// Number of bonded IOBs :         56         56        55
//   IOB Flip Flops :               -          -         -
// Number of BUFGMUXs :             1          1         1
// Number of RAMB16BWEs :           1          1         1
// Clock to Setup 
//on destination clock CLK          ?          ?         ?
// WARNING:Xst:647 - Input <ADDRA<11:10>> is never used. ..
// WARNING:Xst:646 - Signal <r_addrb<11:10>> is assigned but ..

//        ___        ______        ______
//  CLK2X    |______|      |______|      |______
//
//        _____               ______________
//  CLK        |_____________|              |___
//
// Normal                    |
//                         read/write
// CLK2X
//                   |            |
//                 read         write

module rtavr_rom # (
//		parameter SIZE = 1024
		parameter SIZE = 2048
//		parameter SIZE = 4096
) (
`ifdef ROM_DRIVEN_CLK2X
		input CLK2X,
`endif
		input CLK,

		input [11:0] ADDRA,
		output [15:0] DOA,

		input WEB,
`ifdef OUTPUT_MUX_NONE
`else
		input OE,
`endif
		input [12:0] ADDRB,
		input [7:0] DIB, 
		output[7:0] DOB
		);

`ifdef PM_DRIVEN_CLK2X
  wire CLK_A = CLK2X;
`else
  wire CLK_A = CLK;
`endif
`ifdef ROM_DRIVEN_CLK2X
  wire CLK_B = CLK2X;
`else
  wire CLK_B = CLK;
`endif

  reg  [15:0] rom [0:SIZE-1];
  reg  [11:0] r_addra;

  always @(posedge CLK_A)
  begin
`ifdef PM_DRIVEN_CLK2X
  if (~CLK) r_addra <= ADDRA;
`else
	r_addra <= ADDRA;
`endif
  end
  assign DOA = rom[r_addra];

// PORTB 
  reg [11:0] r_addrb;
  reg r_doh;
  reg  [7:0] r_tmp;
  
  wire WEBH = ADDRB[0] & WEB;
  wire WEBL = ~ADDRB[0] & WEB;
  
  always @(posedge CLK_B)
  begin
`ifdef ROM_DRIVEN_CLK2X
	if(WEBL & CLK) r_tmp <=  DIB;
	if(WEBH & CLK) rom[ADDRB[12:1]] <=  { DIB , r_tmp };
	if (~CLK) r_addrb <= ADDRB[12:1];
	if (~CLK) r_doh <= ADDRB[0];
`else
	if(WEBL) r_tmp <=  DIB;
	if(WEBH) rom[ADDRB[12:1]] <=  { DIB , r_tmp };
	r_addrb <= ADDRB[12:1];
	r_doh <= ADDRB[0];
`endif
  end
`ifdef OUTPUT_MUX_NONE
  assign DOB = (r_doh)? rom[r_addrb][15:8]: rom[r_addrb][7:0];
`elsif OUTPUT_MUX_TRI  
  bufif1(DOB, (r_doh)? rom[r_addrb][15:8] : rom[r_addrb][7:0], OE);
`else 
  assign DOB = ~OE ? 8'h00 : (r_doh)? rom[r_addrb][15:8]: rom[r_addrb][7:0];
`endif

//initial
//  begin : rom_data
//    rom[   0] = 16'h0000;
//  end

`ifdef ROM_INCLUDE_V
`include "rom_data.v"
`elsif ROM_INCLUDE_DAT
  initial
    begin 
       $readmemh("rom_data.mem" , rom, 0 , SIZE-1);
    end
`endif

endmodule
