/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if !defined(__linux__) && !defined(_WIN32)

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <string.h>

#else

#include <stdio.h>
#include <malloc.h>
#include <errno.h>
#include <fcntl.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;

#endif

#include "minicrypt.h"

mc_context_t mc_context;

void mc_setkey(const uint8_t *key) {
	if (key) {
		memcpy(mc_context.buf, key, 16);
	} else {
		mc_context.A = 0x67452301;
		mc_context.B = 0xefcdab89;
		mc_context.C = 0x98badcfe;
		mc_context.D = 0x10325476;
	}
}

static void mc_rotate(uint32_t *x) {
	/*  pseudo random generator */
	*x <<= 1;
	if ( (!(*x & (1L<<31)) && (*x & (1L<<28)) )
	     || ((*x & (1L<<31)) && !(*x & (1L<<28))) ) {
		*x |= 1;
	}
	/* netbsd rand() algorism */
	*x = *x * 1103515245 + 12345;
}

static void mc_xor(uint8_t *dst, uint8_t *src) {
	uint8_t i;
	for (i=0; i< 16; i++) {
		*dst = *dst ^ *src;
		dst++;
		src++;
	}
}

void mc_encrypt(uint8_t *block) {
	mc_xor(block,mc_context.buf);
	mc_rotate(&mc_context.A);
	mc_rotate(&mc_context.B);
	mc_rotate(&mc_context.C);
	mc_rotate(&mc_context.D);
	mc_xor(mc_context.buf, block);
}

void mc_decrypt(uint8_t *block) {
	uint8_t tmp[16];
	memcpy(tmp, block, 16);
	mc_xor(block,mc_context.buf);
	mc_rotate(&mc_context.A);
	mc_rotate(&mc_context.B);
	mc_rotate(&mc_context.C);
	mc_rotate(&mc_context.D);
	mc_xor(mc_context.buf, tmp);
}

#if 0
mc_dump(void *p) {
	mc_context_t *mcc;
	int8_t i,j;
	if (p) {
		mcc = (mc_context_t *)p;
	} else {
		mcc = &mc_context;
	}

	//printf("mcc %08x %08x %08x %08x\n",
		mcc->A, mcc->B,
		mcc->C, mcc->D);
}
#endif
