`timescale 10ps / 10ps
`define GTK // use VCD

`include "rtavr_defs.v"

`define MODULE	tb_tmr

module `MODULE;

   reg CLK;  // input
   reg RESET;  // input
   reg INT_ACK;  // input
`ifdef IOR_HAVE_TIM0
   wire IOR_INT0;  // input
`else
   reg IOR_INT0;  // input
`endif
`ifdef IOR_HAVE_TIM0
   wire IOR_TIM0_OVF;  // input
   wire IOR_TIM0_COMPA;  // input
   wire IOR_TIM0_COMPB;  // input
`else
   reg IOR_TIM0_OVF;  // input
   reg IOR_TIM0_COMPA;  // input
   reg IOR_TIM0_COMPB;  // input
`endif

   wire [4:0] INT_VEC;

  reg [7:0] CLK_COUNT;  // input

  rtavr_intcont ICONT (.CLK(CLK), .RESET(RESET)

		, .INT_VEC(INT_VEC)
		, .INT_REQ(INT_REQ)
		, .INT_ACK(INT_ACK)

		, .IOR_INT0(IOR_INT0)
		, .IOR_RST_INT0(IOR_RST_INT0)

`ifdef IOR_HAVE_TIM0
		, .IOR_TIM0_OVF(IOR_TIM0_OVF)
		, .IOR_RST_TIM0_OVF(IOR_RST_TIM0_OVF)
		, .IOR_TIM0_COMPA(IOR_TIM0_COMPA)
		, .IOR_RST_TIM0_COMPA(IOR_RST_TIM0_COMPA)
		, .IOR_TIM0_COMPB(IOR_TIM0_COMPB)
		, .IOR_RST_TIM0_COMPB(IOR_RST_TIM0_COMPB)
`endif
		);
  reg IOR_RD;
  reg IOR_WR;
  reg [5:0] IOR_ADDR;
  reg [7:0] IOR_DI;
  wire [7:0] IOR_DO;

//  `ifdef IOR_HAVE_INT0
//  wire IOR_RST_INT0;
//  `endif
//  `ifdef IOR_HAVE_TIM0
//  wire IOR_RST_TIM0_OVF;
//  wire IOR_RST_TIM0_COMPA;
//  wire IOR_RST_TIM0_COMPB;
//  `endif

  rtavr_ior IOR (.CLK(CLK), .RESET(RESET), .WR(IOR_WR) , .RD(IOR_RD)
		, .ADDR(IOR_ADDR[5:0]), .DI(IOR_DI[7:0]), .DO(IOR_DO[7:0])
  `ifdef IOR_HAVE_PORTA
		, .PORTA(PORTA[7:0])
  `endif
  `ifdef IOR_HAVE_PORTB
		, .PORTB(PORTB[7:0])
  `endif
  `ifdef IOR_HAVE_PORTC
		, .PORTC(PORTC[7:0])
  `endif
  `ifdef IOR_HAVE_INT0
		, .INT0(IOR_INT0)
		, .RST_INT0(IOR_RST_INT0)
  `endif
  `ifdef IOR_HAVE_TIM0
		// INTERRUPTS 
		, .TIM0_OVF(IOR_TIM0_OVF)
		, .RST_TIM0_OVF(IOR_RST_TIM0_OVF)
		, .TIM0_COMPA(IOR_TIM0_COMPA)
		, .RST_TIM0_COMPA(IOR_RST_TIM0_COMPA)
		, .TIM0_COMPB(IOR_TIM0_COMPB)
		, .RST_TIM0_COMPB(IOR_RST_TIM0_COMPB)
  `endif
		);

   parameter CLKP = 5000;        //  50ns
   parameter CLKH = CLKP/2;      //
   parameter CLKL = CLKP-CLKH;   //
   
   always begin
      #CLKH CLK = 0;
      #CLKL CLK = 1;
   end

`ifdef GTK
   initial begin
      $dumpfile("tb.vcd");
      $dumpvars(0,`MODULE);
   end
`endif

   initial begin : test
      sys_reset();
      ior_out(6'h18, 8'h10); // TSM = 0, PSR = 1
      sys_wait(4);
      ior_out(6'h18, 8'h30); // TSM = 1, PSR = 1
      sys_wait(4);
      ior_out(6'h18, 8'h00); // TSM = 0, PSR = 0
      ior_out(6'h18, 8'h01); // TSM = 0, PSR = 0 , CS = 1 (1/1)
      sys_wait(4);
      ior_out(6'h18, 8'h12); // TSM = 0, PSR = 1 , CS = 2 (1/8)
      sys_wait(20);
      ior_out(6'h26, 8'h07); // OCIE0B, OCIE0A = 1, TOIE0 = 1 
      ior_out(6'h18, 8'h01); // TSM = 0, PSR = 0 , CS = 1 (1/1)
      ior_out(6'h15, 8'hfb); // OCR0B = 0xfb
      ior_out(6'h16, 8'hfd); // OCR0A = 0xfd
      ior_out(6'h17, 8'hf8); // TCNT0 = 0xf8
      sys_wait_int(20);
      sys_wait_int(20);
      sys_wait_int(20);
      sys_wait(20);
      $display("*** Simulation Complete! ***");
      $finish;
   end

   task sys_reset;
      begin
	 INT_ACK = 0;
	 IOR_RD = 0;
	 IOR_WR = 0;
	 IOR_ADDR[5:0] = 0;
         IOR_DI[7:0] =0;
`ifdef IOR_HAVE_TIM0
`else
   	 IOR_INT0 = 0;
`endif
`ifdef IOR_HAVE_TIM0
`else
	 IOR_TIM0_OVF = 0;
	 IOR_TIM0_COMPA = 0;
	 IOR_TIM0_COMPB = 0;
`endif
         #1 RESET = 0;
         CLK = 0;
	 @(negedge CLK);
         #1 RESET = 1;
         @(posedge CLK);
      end
   endtask // sys_reset
   
   task ior_out;
      input [5:0] addr;
      input [7:0] data;
      begin
	 @(negedge CLK);
         #1 IOR_ADDR[5:0] = addr[5:0];
         IOR_DI[7:0] = data[7:0];
         IOR_WR = 1;
         @(posedge CLK);
	 @(negedge CLK);
	 #1 IOR_ADDR[5:0] = 0;
         IOR_DI[7:0] =0;
         IOR_WR = 0;
         @(posedge CLK);
      end
   endtask // ior_out

   task ior_in;
      input [5:0] addr;
      begin
	 @(negedge CLK);
         #1 IOR_ADDR[5:0] = addr[5:0];
         IOR_RD = 1;
         @(posedge CLK);
	 @(negedge CLK);
         #1 IOR_ADDR[5:0] = 0;
         IOR_RD = 0;
         @(posedge CLK);
      end
   endtask // ior_out

   task sys_wait;
      input [7:0] count;
      begin
        CLK_COUNT = 0;
        while ( CLK_COUNT < count )
        begin
           @(posedge CLK);
           CLK_COUNT = CLK_COUNT + 1;
        end
      end
   endtask // sys_wait

   task sys_wait_int;
      input [7:0] count;
      begin
        CLK_COUNT = 0;
        while ( ( CLK_COUNT < count ) & ~INT_REQ )
        begin
           @(posedge CLK);
           CLK_COUNT = CLK_COUNT + 1;
        end
        @(negedge CLK);
	INT_ACK = 1;
        @(negedge CLK);
        @(negedge CLK);
        @(negedge CLK);
	INT_ACK = 0;
        @(posedge CLK);
        @(posedge CLK);
      end
   endtask // sys_wait

endmodule // test
